"""Fixtures for tests."""
import pathlib
import typing as t

import docutils.parsers.rst
import pytest

from rstcheck import _extras


if _extras.SPHINX_INSTALLED:
    import sphinx.application


REPO_DIR = pathlib.Path(__file__).resolve().parents[1].resolve()
TESTING_DIR = REPO_DIR / "testing"
EXAMPLES_DIR = TESTING_DIR / "examples"


@pytest.fixture(name="patch_docutils_directives_and_roles_dict")
def _patch_docutils_directives_and_roles_dict_fixture(monkeypatch: pytest.MonkeyPatch) -> None:
    """Monkeypatch docutils' directives and roles state dicts.

    This patch is required when tests are run in parallel (default), because they would
    under the hood all write to the same state dicts otherwise and influence each other.
    """
    test_dict_directives: t.Dict[str, t.Any] = {}
    test_dict_roles: t.Dict[str, t.Any] = {}

    if _extras.SPHINX_INSTALLED:
        monkeypatch.setattr(
            sphinx.application.docutils.directives, "_directives", test_dict_directives
        )
        monkeypatch.setattr(sphinx.application.docutils.roles, "_roles", test_dict_roles)
    else:
        monkeypatch.setattr(docutils.parsers.rst.directives, "_directives", test_dict_directives)
        monkeypatch.setattr(docutils.parsers.rst.roles, "_roles", test_dict_roles)



"""Tests for ``config`` module."""
# pylint: disable=too-many-lines
import pathlib
import re
import typing as t

import pytest

from rstcheck import _extras, config


def test_report_level_map_matches_numbers() -> None:  # noqa: AAA01
    """Test that the enum's values match the map's ones."""
    enum_values = [e.value for e in config.ReportLevel]
    map_values = list(config.ReportLevelMap.values())

    assert map_values == enum_values


def test_report_level_map_matches_names() -> None:  # noqa: AAA01
    """Test that the enum's name match the map's keys."""
    enum_names = [
        e.casefold()
        for e in config.ReportLevel._member_names_  # pylint: disable=protected-access,no-member
    ]
    map_keys = list(config.ReportLevelMap.keys())

    assert enum_names == map_keys


def test_default_values_for_config_file() -> None:
    """Test default values of  ``RstcheckConfigFile``."""
    result = config.RstcheckConfigFile()

    assert result.report_level is None
    assert result.ignore_directives is None
    assert result.ignore_roles is None
    assert result.ignore_substitutions is None
    assert result.ignore_languages is None
    assert result.ignore_messages is None


def test_default_values_for_config() -> None:
    """Test default values of  ``RstcheckConfig``."""
    result = config.RstcheckConfig()

    assert result.report_level is None
    assert result.ignore_directives is None
    assert result.ignore_roles is None
    assert result.ignore_substitutions is None
    assert result.ignore_languages is None
    assert result.ignore_messages is None
    assert result.config_path is None
    assert result.recursive is None


class TestReportLevelValidator:
    """Test ``valid_report_level`` validator method of the ``RstcheckConfig`` class.

    It validates the ``report_level`` setting.
    """

    @staticmethod
    def test_none_means_none() -> None:
        """Test ``None`` results in no report level."""
        result = config.RstcheckConfigFile(report_level=None)

        assert result is not None
        assert result.report_level is None

    @staticmethod
    def test_set_level_stays() -> None:
        """Test set level results in same level."""
        result = config.RstcheckConfigFile(report_level=config.ReportLevel.SEVERE)

        assert result is not None
        assert result.report_level is config.ReportLevel.SEVERE

    @staticmethod
    def test_empty_string_means_default() -> None:
        """Test empty string results in default report level."""
        result = config.RstcheckConfigFile(report_level="")

        assert result is not None
        assert result.report_level is config.ReportLevel.INFO

    @staticmethod
    @pytest.mark.parametrize(
        "level",
        [1, 2, 3, 4, 5, "info", "warning", "error", "severe", "none", "NONE", "None", "NoNe"],
    )
    def test_valid_report_levels(level: t.Any) -> None:  # noqa: ANN401
        """Test valid report levels accepted by docutils."""
        result = config.RstcheckConfigFile(report_level=level)

        assert result is not None
        assert result.report_level is not None

    @staticmethod
    @pytest.mark.parametrize(
        "level",
        [-1, 0, 1.5, 6, 32, False, True, "information", "warn", "err", "critical", "fatal"],
    )
    def test_invalid_report_levels(level: t.Any) -> None:  # noqa: ANN401
        """Test invalid report levels not accepted by docutils."""
        with pytest.raises(ValueError, match="Invalid report level"):
            config.RstcheckConfigFile(report_level=level)


class TestSplitStrValidator:
    """Test ``split_str`` validator method of the ``RstcheckConfig`` class.

    It validates the

    - ``ignore_directives``
    - ``ignore_roles``
    - ``ignore_substitutions``
    - ``ignore_languages``

    settings.
    """

    @staticmethod
    def test_none_means_default() -> None:
        """Test ``None`` results in unset ignore_messages."""
        result = config.RstcheckConfigFile(
            ignore_languages=None,
            ignore_directives=None,
            ignore_roles=None,
            ignore_substitutions=None,
        )

        assert result is not None
        assert result.ignore_languages is None
        assert result.ignore_directives is None
        assert result.ignore_roles is None
        assert result.ignore_substitutions is None

    @staticmethod
    @pytest.mark.parametrize(
        ("string", "split_list"),
        [
            ("value1", ["value1"]),
            ("value1,value2", ["value1", "value2"]),
            ("value1, value2", ["value1", "value2"]),
            ("value1 ,value2", ["value1", "value2"]),
            ("value1 , value2", ["value1", "value2"]),
            ("value1 ,\n value2", ["value1", "value2"]),
            ("value1 ,\n value2\n", ["value1", "value2"]),
            ("value1 , value2,", ["value1", "value2"]),
            ("value1 , value2 ,", ["value1", "value2"]),
            ("value1 , value2 , ", ["value1", "value2"]),
        ],
    )
    def test_strings_are_transformed_to_lists(string: str, split_list: t.List[str]) -> None:
        """Test strings are split at the ",", trailing commas are ignored and whitespace cleaned."""
        result = config.RstcheckConfigFile(
            ignore_languages=string,
            ignore_directives=string,
            ignore_roles=string,
            ignore_substitutions=string,
        )

        assert result is not None
        assert result.ignore_languages == split_list
        assert result.ignore_directives == split_list
        assert result.ignore_roles == split_list
        assert result.ignore_substitutions == split_list

    @staticmethod
    @pytest.mark.parametrize(
        ("string_list", "string_list_cleaned"),
        [
            (["value1"], ["value1"]),
            (["value1", "value2"], ["value1", "value2"]),
            (["value1", " value2"], ["value1", "value2"]),
            (["value1 ", "value2"], ["value1", "value2"]),
            (["value1 ", " value2"], ["value1", "value2"]),
        ],
    )
    def test_string_lists_are_whitespace_cleaned(
        string_list: t.List[str], string_list_cleaned: t.List[str]
    ) -> None:
        """Test lists of strings are whitespace cleaned."""
        result = config.RstcheckConfigFile(
            ignore_languages=string_list,
            ignore_directives=string_list,
            ignore_roles=string_list,
            ignore_substitutions=string_list,
        )

        assert result is not None
        assert result.ignore_languages == string_list_cleaned
        assert result.ignore_directives == string_list_cleaned
        assert result.ignore_roles == string_list_cleaned
        assert result.ignore_substitutions == string_list_cleaned

    @staticmethod
    @pytest.mark.parametrize(
        "value",
        [
            1,
            [1],
            1.1,
            [1.1],
            False,
            [False],
            True,
            [True],
            ["foo", 1],
        ],
    )
    def test_invalid_settings(value: str) -> None:
        """Test invalid settings."""
        with pytest.raises(ValueError, match="Not a string or list of strings"):
            config.RstcheckConfigFile(
                ignore_languages=value,
                ignore_directives=value,
                ignore_roles=value,
                ignore_substitutions=value,
            )


class TestJoinRegexStrValidator:
    """Test ``join_regex_str`` validator method of the ``RstcheckConfig`` class.

    It validates the ``ignore_messages`` setting.
    """

    @staticmethod
    def test_none_means_unset() -> None:
        """Test ``None`` results in unset ignore_messages."""
        result = config.RstcheckConfigFile(ignore_messages=None)

        assert result is not None
        assert result.ignore_messages is None

    @staticmethod
    def test_passed_regex_is_kept() -> None:
        """Test passing a regex."""
        result = config.RstcheckConfigFile(ignore_messages=re.compile(""))

        assert result is not None
        assert isinstance(result.ignore_messages, re.Pattern)

    @staticmethod
    def test_strings_are_parsed_as_regex() -> None:
        """Test strings are parsed as regex."""
        string = r"\d{4}\.[A-Z]+Test$"
        regex = re.compile(string)

        result = config.RstcheckConfigFile(ignore_messages=string)

        assert result is not None
        assert result.ignore_messages == regex

    @staticmethod
    def test_empty_strings_are_valid() -> None:
        """Test empty strings are parsed as regex too."""
        string = ""
        regex = re.compile(string)

        result = config.RstcheckConfigFile(ignore_messages=string)

        assert result is not None
        assert result.ignore_messages == regex

    @staticmethod
    def test_string_list_are_joined_and_parsed_as_regex() -> None:
        """Test ignore_messages string lists are joined with "|" and parsed as regex."""
        string_list = [r"\d{4}\.[A-Z]+Test$", r"\d{4}\.[A-Z]+Test2$", r"\d{4}\.[A-Z]+Test3$"]
        full_string = r"\d{4}\.[A-Z]+Test$|\d{4}\.[A-Z]+Test2$|\d{4}\.[A-Z]+Test3$"
        regex = re.compile(full_string)

        result = config.RstcheckConfigFile(ignore_messages=string_list)

        assert result is not None
        assert result.ignore_messages == regex

    @staticmethod
    @pytest.mark.parametrize(
        ("string_list", "full_string"), [([""], ""), (["", ""], "|"), ([], "")]
    )
    def test_list_with_empty_contents(string_list: t.List[str], full_string: str) -> None:
        """Test list with empty contents are parsed as regex too."""
        regex = re.compile(full_string)

        result = config.RstcheckConfigFile(ignore_messages=string_list)

        assert result is not None
        assert result.ignore_messages == regex

    @staticmethod
    @pytest.mark.parametrize(
        "value",
        [
            1,
            [1],
            1.1,
            [1.1],
            False,
            [False],
            True,
            [True],
            ["foo", 1],
        ],
    )
    def test_invalid_settings(value: str) -> None:
        """Test invalid ignore_messages settings."""
        with pytest.raises(ValueError, match="Not a string or list of strings"):
            config.RstcheckConfigFile(ignore_messages=value)


class TestIniFileLoader:
    """Test ``load_config_from_ini_file``."""

    @staticmethod
    def test_missing_file_errors(tmp_path: pathlib.Path) -> None:
        """Test ``FileNotFoundError`` is raised on missing file."""
        conf_file = tmp_path / "config.ini"

        with pytest.raises(FileNotFoundError):
            config._load_config_from_ini_file(conf_file)  # pylint: disable=protected-access

    @staticmethod
    def test_not_a_file_errors(tmp_path: pathlib.Path) -> None:
        """Test ``FileNotFoundError`` is raised when not file."""
        conf_file = tmp_path

        with pytest.raises(FileNotFoundError):
            config._load_config_from_ini_file(conf_file)  # pylint: disable=protected-access

    @staticmethod
    def test_returns_none_on_missing_section(tmp_path: pathlib.Path) -> None:
        """Test ``None`` is returned on missing section."""
        conf_file = tmp_path / "config.ini"
        file_content = "[not-rstcheck]"
        conf_file.write_text(file_content)

        result = config._load_config_from_ini_file(conf_file)  # pylint: disable=protected-access

        assert result is None

    @staticmethod
    def test_ignores_unsupported_settings(tmp_path: pathlib.Path) -> None:
        """Test unsupported settings are ignored."""
        conf_file = tmp_path / "config.ini"
        file_content = """[rstcheck]
        unsupported_feature=True
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_ini_file(conf_file)  # pylint: disable=protected-access

        assert result is not None

    @staticmethod
    def test_supported_settings_are_loaded(tmp_path: pathlib.Path) -> None:
        """Test supported settings are loaded."""
        conf_file = tmp_path / "config.ini"
        file_content = """[rstcheck]
        report_level=3
        ignore_directives=directive
        ignore_roles=role
        ignore_substitutions=substitution
        ignore_languages=language
        ignore_messages=message
        """
        conf_file.write_text(file_content)
        regex = re.compile("message")

        result = config._load_config_from_ini_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR
        assert result.ignore_directives == ["directive"]
        assert result.ignore_roles == ["role"]
        assert result.ignore_substitutions == ["substitution"]
        assert result.ignore_languages == ["language"]
        assert result.ignore_messages == regex

    @staticmethod
    def test_multiline_strings(tmp_path: pathlib.Path) -> None:
        """Test multiline strings are parsed."""
        conf_file = tmp_path / "config.ini"
        file_content = """[rstcheck]
        ignore_directives=directive
        ignore_roles=
            role1,
            role2,
            role3
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_ini_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.ignore_directives == ["directive"]
        assert result.ignore_roles == ["role1", "role2", "role3"]

    @staticmethod
    def test_multiline_strings_trailing_comma(tmp_path: pathlib.Path) -> None:
        """Test multiline strings with trailing comma are parsed."""
        conf_file = tmp_path / "config.ini"
        file_content = """[rstcheck]
        ignore_directives=directive
        ignore_roles=
            role1,
            role2,
            role3,
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_ini_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.ignore_directives == ["directive"]
        assert result.ignore_roles == ["role1", "role2", "role3"]

    @staticmethod
    def test_file_with_mixed_supported_settings(tmp_path: pathlib.Path) -> None:
        """Test mix of supported and unsupported settings."""
        conf_file = tmp_path / "config.ini"
        file_content = """[rstcheck]
        report_level=3
        ignore_directives=directive
        unsupported_feature=True
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_ini_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR
        assert result.ignore_directives == ["directive"]

    @staticmethod
    def test_file_with_mixed_supported_sections(tmp_path: pathlib.Path) -> None:
        """Test mix of rstcheck and other section."""
        conf_file = tmp_path / "config.ini"
        file_content = """[rstcheck]
        report_level=3
        ignore_directives=directive
        unsupported_feature=True

        [not-rstcheck]
        report_level=2
        ignore_directives=not-directive
        unsupported_feature=False
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_ini_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR
        assert result.ignore_directives == ["directive"]


@pytest.mark.skipif(not _extras.TOMLI_INSTALLED, reason="Depends on toml extra.")
class TestTomlFileLoader:
    """Test ``load_config_from_toml_file``."""

    @staticmethod
    def test_wrong_file_suffix_errors(tmp_path: pathlib.Path) -> None:
        """Test ``ValueError`` is raised on wrong file suffix."""
        conf_file = tmp_path / "config.ini"
        conf_file.touch()

        with pytest.raises(ValueError, match="File is not a TOML file"):
            config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

    @staticmethod
    def test_missing_file_errors(tmp_path: pathlib.Path) -> None:
        """Test ``FileNotFoundError`` is raised on missing file."""
        conf_file = tmp_path / "config.toml"

        with pytest.raises(FileNotFoundError):
            config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

    @staticmethod
    def test_not_a_file_errors(tmp_path: pathlib.Path) -> None:
        """Test ``FileNotFoundError`` is raised when not file."""
        conf_file = tmp_path / "config.toml"
        conf_file.mkdir()

        with pytest.raises(FileNotFoundError):
            config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

    @staticmethod
    @pytest.mark.parametrize("invalid_section", ["[tool.not-rstcheck]", "[rstcheck]"])
    def test_returns_none_on_missing_section(tmp_path: pathlib.Path, invalid_section: str) -> None:
        """Test ``None`` is returned on missing section."""
        conf_file = tmp_path / "config.toml"
        conf_file.write_text(invalid_section)

        result = config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

        assert result is None

    @staticmethod
    def test_ignores_unsupported_settings(tmp_path: pathlib.Path) -> None:
        """Test unsupported settings are ignored."""
        conf_file = tmp_path / "config.toml"
        file_content = """[tool.rstcheck]
        unsupported_feature = true
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

        assert result is not None

    @staticmethod
    def test_supported_settings_are_loaded(tmp_path: pathlib.Path) -> None:
        """Test supported settings are loaded."""
        conf_file = tmp_path / "config.toml"
        file_content = """[tool.rstcheck]
        report_level = 3
        ignore_directives = ["directive"]
        ignore_roles = ["role"]
        ignore_substitutions = ["substitution"]
        ignore_languages = ["language"]
        ignore_messages = "message"
        """
        conf_file.write_text(file_content)
        regex = re.compile("message")

        result = config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR
        assert result.ignore_directives == ["directive"]
        assert result.ignore_roles == ["role"]
        assert result.ignore_substitutions == ["substitution"]
        assert result.ignore_languages == ["language"]
        assert result.ignore_messages == regex

    @staticmethod
    def test_file_with_mixed_supported_settings(tmp_path: pathlib.Path) -> None:
        """Test mix of supported and unsupported settings."""
        conf_file = tmp_path / "config.toml"
        file_content = """[tool.rstcheck]
        report_level = 3
        ignore_directives = ["directive"]
        unsupported_feature = true
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR
        assert result.ignore_directives == ["directive"]

    @staticmethod
    def test_file_with_mixed_supported_sections(tmp_path: pathlib.Path) -> None:
        """Test mix of rstcheck and other section."""
        conf_file = tmp_path / "config.toml"
        file_content = """[tool.rstcheck]
        report_level = 3
        ignore_directives = ["directive"]
        unsupported_feature = true

        [tool.not-rstcheck]
        report_level = 2
        ignore_directives = "not-directive"
        unsupported_feature = false
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR
        assert result.ignore_directives == ["directive"]

    @staticmethod
    @pytest.mark.parametrize(
        ("value", "parsed_value"),
        [
            ("none", config.ReportLevel.NONE),
            ("ERROR", config.ReportLevel.ERROR),
            ("1", config.ReportLevel.INFO),
            ("3", config.ReportLevel.ERROR),
        ],
    )
    def test_report_level_as_strings(
        tmp_path: pathlib.Path, value: str, parsed_value: config.ReportLevel
    ) -> None:
        """Test report setting with string values."""
        conf_file = tmp_path / "config.toml"
        file_content = f"""[tool.rstcheck]
        report_level = "{value}"
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.report_level == parsed_value

    @staticmethod
    @pytest.mark.parametrize(
        ("value", "parsed_value"),
        [
            (1, config.ReportLevel.INFO),
            (3, config.ReportLevel.ERROR),
            (5, config.ReportLevel.NONE),
        ],
    )
    def test_report_level_as_int(
        tmp_path: pathlib.Path, value: int, parsed_value: config.ReportLevel
    ) -> None:
        """Test report setting with integer values."""
        conf_file = tmp_path / "config.toml"
        file_content = f"""[tool.rstcheck]
        report_level = {value}
        """
        conf_file.write_text(file_content)

        result = config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.report_level == parsed_value

    @staticmethod
    def test_ignore_messages_as_str(tmp_path: pathlib.Path) -> None:
        """Test ignore_messages setting with string value."""
        conf_file = tmp_path / "config.toml"
        file_content = """[tool.rstcheck]
        ignore_messages = "some-regex"
        """
        conf_file.write_text(file_content)
        regex = re.compile("some-regex")

        result = config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.ignore_messages == regex

    @staticmethod
    def test_ignore_messages_as_list(tmp_path: pathlib.Path) -> None:
        """Test ignore_messages setting with list of strings value."""
        conf_file = tmp_path / "config.toml"
        file_content = """[tool.rstcheck]
        ignore_messages = ["some-regex", "another-regex"]
        """
        conf_file.write_text(file_content)
        regex = re.compile(r"some-regex|another-regex")

        result = config._load_config_from_toml_file(conf_file)  # pylint: disable=protected-access

        assert result is not None
        assert result.ignore_messages == regex


class TestConfigFileLoader:
    """Test ``load_config_file``."""

    @staticmethod
    @pytest.mark.parametrize("ini_file", [".rstcheck.cfg", "setup.cfg", "config.ini", "config.cfg"])
    def test_ini_files(tmp_path: pathlib.Path, ini_file: str) -> None:
        """Test with INI files."""
        conf_file = tmp_path / ini_file
        file_content = """[rstcheck]
        report_level = 3
        """
        conf_file.write_text(file_content)

        result = config.load_config_file(conf_file)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    @pytest.mark.skipif(not _extras.TOMLI_INSTALLED, reason="Depends on toml extra.")
    @pytest.mark.parametrize("toml_file", [".rstcheck.toml", "setup.toml", "config.toml"])
    def test_toml_files(tmp_path: pathlib.Path, toml_file: str) -> None:
        """Test with TOML files."""
        conf_file = tmp_path / toml_file
        file_content = """[tool.rstcheck]
        report_level = 3
        """
        conf_file.write_text(file_content)

        result = config.load_config_file(conf_file)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR


class TestConfigDirLoader:
    """Test ``load_config_file_from_dir``."""

    @staticmethod
    @pytest.mark.parametrize("ini_file", [".rstcheck.cfg", "setup.cfg"])
    def test_supported_ini_files(tmp_path: pathlib.Path, ini_file: str) -> None:
        """Test with supported INI files."""
        conf_file = tmp_path / ini_file
        file_content = """[rstcheck]
        report_level = 3
        """
        conf_file.write_text(file_content)

        result = config.load_config_file_from_dir(tmp_path)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    @pytest.mark.parametrize("ini_file", ["config.ini", "config.cfg"])
    def test_unsupported_ini_files(tmp_path: pathlib.Path, ini_file: str) -> None:
        """Test with unsupported INI files."""
        conf_file = tmp_path / ini_file
        file_content = """[rstcheck]
        report_level = 3
        """
        conf_file.write_text(file_content)

        result = config.load_config_file_from_dir(tmp_path)

        assert result is None

    @staticmethod
    @pytest.mark.skipif(not _extras.TOMLI_INSTALLED, reason="Depends on toml extra.")
    @pytest.mark.parametrize("toml_file", ["pyproject.toml"])
    def test_supported_toml_files(tmp_path: pathlib.Path, toml_file: str) -> None:
        """Test with supported TOML files."""
        conf_file = tmp_path / toml_file
        file_content = """[tool.rstcheck]
        report_level = 3
        """
        conf_file.write_text(file_content)

        result = config.load_config_file_from_dir(tmp_path)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    @pytest.mark.skipif(not _extras.TOMLI_INSTALLED, reason="Depends on toml extra.")
    @pytest.mark.parametrize("toml_file", [".rstcheck.toml", "setup.toml", "config.toml"])
    def test_unsupported_toml_files(tmp_path: pathlib.Path, toml_file: str) -> None:
        """Test with unsupported TOML files."""
        conf_file = tmp_path / toml_file
        file_content = """[tool.rstcheck]
        report_level = 3
        """
        conf_file.write_text(file_content)

        result = config.load_config_file_from_dir(tmp_path)

        assert result is None

    @staticmethod
    def test_rstcheck_over_setup(tmp_path: pathlib.Path) -> None:
        """Test .rstcheck.cfg takes precedence over setup.cfg."""
        setup_conf_file = tmp_path / "setup.cfg"
        setup_file_content = """[rstcheck]
        report_level = 2
        """
        setup_conf_file.write_text(setup_file_content)
        rstcheck_conf_file = tmp_path / ".rstcheck.cfg"
        rstcheck_file_content = """[rstcheck]
        report_level = 3
        """
        rstcheck_conf_file.write_text(rstcheck_file_content)

        result = config.load_config_file_from_dir(tmp_path)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    @pytest.mark.skipif(not _extras.TOMLI_INSTALLED, reason="Depends on toml extra.")
    def test_rstcheck_over_pyproject(tmp_path: pathlib.Path) -> None:
        """Test .rstcheck.cfg takes precedence over pyproject.toml."""
        pyproject_conf_file = tmp_path / "pyproject.toml"
        pyproject_file_content = """[tool.rstcheck]
        report_level = 2
        """
        pyproject_conf_file.write_text(pyproject_file_content)
        rstcheck_conf_file = tmp_path / ".rstcheck.cfg"
        rstcheck_file_content = """[rstcheck]
        report_level = 3
        """
        rstcheck_conf_file.write_text(rstcheck_file_content)

        result = config.load_config_file_from_dir(tmp_path)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    @pytest.mark.skipif(not _extras.TOMLI_INSTALLED, reason="Depends on toml extra.")
    def test_pyproject_over_setup(tmp_path: pathlib.Path) -> None:
        """Test pyproject.toml takes precedence over setup.cfg."""
        setup_conf_file = tmp_path / "setup.cfg"
        setup_file_content = """[rstcheck]
        report_level = 2
        """
        setup_conf_file.write_text(setup_file_content)
        pyproject_conf_file = tmp_path / "pyproject.toml"
        pyproject_file_content = """[tool.rstcheck]
        report_level = 3
        """
        pyproject_conf_file.write_text(pyproject_file_content)

        result = config.load_config_file_from_dir(tmp_path)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    @pytest.mark.skipif(not _extras.TOMLI_INSTALLED, reason="Depends on toml extra.")
    def test_missing_config_means_next_file_is_checked(tmp_path: pathlib.Path) -> None:
        """Test missing config in file results in checking of next file."""
        setup_conf_file = tmp_path / "setup.cfg"
        setup_file_content = """[rstcheck]
        report_level = 3
        """
        setup_conf_file.write_text(setup_file_content)
        rstcheck_conf_file = tmp_path / ".rstcheck.cfg"
        rstcheck_file_content = """[not-rstcheck]
        report_level = 2
        """
        rstcheck_conf_file.write_text(rstcheck_file_content)

        result = config.load_config_file_from_dir(tmp_path)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR


class TestConfigDirTreeLoader:
    """Test ``load_config_file_from_dir_tree``."""

    @staticmethod
    def test_parent_searching(tmp_path: pathlib.Path) -> None:
        """Test option to search up the dir tree."""
        nested_dir = tmp_path / "nested"
        nested_dir.mkdir()
        unsupported_file = nested_dir / "config.cfg"
        unsupported_file_content = """[rstcheck]
        report_level = 2
        """
        unsupported_file.write_text(unsupported_file_content)
        supported_file = tmp_path / "setup.cfg"
        supported_file_content = """[rstcheck]
        report_level = 3
        """
        supported_file.write_text(supported_file_content)

        result = config.load_config_file_from_dir_tree(nested_dir)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    def test_no_file_up_to_root(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
        """Test option to search up the dir tree with no file up to root dir."""
        root_dir = tmp_path
        nested_dir = tmp_path / "nested"
        nested_dir.mkdir()
        monkeypatch.setattr(pathlib.Path, "parent", root_dir)

        result = config.load_config_file_from_dir_tree(nested_dir)

        assert result is None


class TestConfigPathLoader:
    """Test ``load_config_file_from_path``."""

    @staticmethod
    def test_with_nonexisting_path() -> None:
        """Test with INI file."""
        conf_file = pathlib.Path("does-not-exist-cfg")

        result = config.load_config_file_from_path(conf_file)

        assert result is None

    @staticmethod
    def test_with_file(tmp_path: pathlib.Path) -> None:
        """Test with INI file."""
        conf_file = tmp_path / ".rstcheck.cfg"
        file_content = """[rstcheck]
        report_level = 3
        """
        conf_file.write_text(file_content)

        result = config.load_config_file_from_path(conf_file)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    def test_with_dir(tmp_path: pathlib.Path) -> None:
        """Test with directory."""
        conf_file = tmp_path / ".rstcheck.cfg"
        file_content = """[rstcheck]
        report_level = 3
        """
        conf_file.write_text(file_content)

        result = config.load_config_file_from_path(tmp_path)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    def test_with_nested_dir(tmp_path: pathlib.Path) -> None:
        """Test with nested dir tree."""
        nested_dir = tmp_path / "nested"
        nested_dir.mkdir()
        conf_file = tmp_path / "setup.cfg"
        file_content = """[rstcheck]
        report_level = 3
        """
        conf_file.write_text(file_content)

        result = config.load_config_file_from_path(nested_dir, search_dir_tree=True)

        assert result is not None
        assert result.report_level == config.ReportLevel.ERROR


class TestConfigMerger:
    """Test ``merge_configs``."""

    @staticmethod
    def test_default_merge_with_full_config() -> None:
        """Test config merging with full config."""
        config_base = config.RstcheckConfig(report_level=config.ReportLevel.SEVERE)
        config_add = config.RstcheckConfig(report_level=config.ReportLevel.ERROR)

        result = config.merge_configs(config_base, config_add)

        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    def test_default_merge_with_file_config() -> None:
        """Test config merging with file config."""
        config_base = config.RstcheckConfig(report_level=config.ReportLevel.SEVERE)
        config_add = config.RstcheckConfigFile(report_level=config.ReportLevel.ERROR)

        result = config.merge_configs(config_base, config_add)

        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    def test_default_merge_with_full_config_and_empty_base_config() -> None:
        """Test config merging with full config and empty base config."""
        config_base = config.RstcheckConfig()
        config_add = config.RstcheckConfig(report_level=config.ReportLevel.ERROR)

        result = config.merge_configs(config_base, config_add)

        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    def test_default_merge_with_file_config_and_empty_base_config() -> None:
        """Test config merging with file config and empty base config."""
        config_base = config.RstcheckConfig()
        config_add = config.RstcheckConfigFile(report_level=config.ReportLevel.ERROR)

        result = config.merge_configs(config_base, config_add)

        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    def test_default_merge_with_full_config_and_changed_dominance() -> None:
        """Test config merging with full config and changed dominance."""
        config_base = config.RstcheckConfig(report_level=config.ReportLevel.SEVERE)
        config_add = config.RstcheckConfig(report_level=config.ReportLevel.ERROR)

        result = config.merge_configs(config_base, config_add, config_add_is_dominant=False)

        assert result.report_level == config.ReportLevel.SEVERE

    @staticmethod
    def test_default_merge_with_file_config_and_changed_dominance() -> None:
        """Test config merging with file config and changed dominance."""
        config_base = config.RstcheckConfig(report_level=config.ReportLevel.SEVERE)
        config_add = config.RstcheckConfigFile(report_level=config.ReportLevel.ERROR)

        result = config.merge_configs(config_base, config_add, config_add_is_dominant=False)

        assert result.report_level == config.ReportLevel.SEVERE

    @staticmethod
    def test_default_merge_with_full_config_and_changed_dominance_and_empty_base_config() -> None:
        """Test config merging with full config, changed dominance and empty base config.

        Test ``None`` gets overwritten by set value.
        """
        config_base = config.RstcheckConfig()
        config_add = config.RstcheckConfig(report_level=config.ReportLevel.ERROR)

        result = config.merge_configs(config_base, config_add, config_add_is_dominant=False)

        assert result.report_level == config.ReportLevel.ERROR

    @staticmethod
    def test_default_merge_with_file_config_and_changed_dominance_and_empty_base_config() -> None:
        """Test config merging with file config, changed dominance and empty base config.

        Test ``None`` gets overwritten by set value.
        """
        config_base = config.RstcheckConfig()
        config_add = config.RstcheckConfigFile(report_level=config.ReportLevel.ERROR)

        result = config.merge_configs(config_base, config_add, config_add_is_dominant=False)

        assert result.report_level == config.ReportLevel.ERROR



"""Rstcheck configuration functionality."""
import configparser
import contextlib
import enum
import pathlib
import re
import typing as t

import pydantic

from . import _extras


if _extras.TOMLI_INSTALLED:  # pragma: no cover
    import tomli


CONFIG_FILES = [".rstcheck.cfg", "setup.cfg"]
if _extras.TOMLI_INSTALLED:  # pragma: no cover
    CONFIG_FILES = [".rstcheck.cfg", "pyproject.toml", "setup.cfg"]


class ReportLevel(enum.Enum):
    """Report levels supported by docutils."""

    INFO = 1
    WARNING = 2
    ERROR = 3
    SEVERE = 4
    NONE = 5


ReportLevelMap = {
    "info": 1,
    "warning": 2,
    "error": 3,
    "severe": 4,
    "none": 5,
}
"""Map docutils report levels in text form to numbers."""


DEFAULT_REPORT_LEVEL = ReportLevel.INFO


class RstcheckConfigFile(pydantic.BaseModel):  # pylint: disable=no-member
    """Rstcheck config file.

    :raises ValueError:: If setting has incorrect value or type
    :raises pydantic.error_wrappers.ValidationError:: If setting is not parsable into correct type
    """

    report_level: t.Optional[ReportLevel]
    ignore_directives: t.Optional[t.List[str]]
    ignore_roles: t.Optional[t.List[str]]
    ignore_substitutions: t.Optional[t.List[str]]
    ignore_languages: t.Optional[t.List[str]]
    # NOTE: Pattern type-arg errors pydanic: https://github.com/samuelcolvin/pydantic/issues/2636
    ignore_messages: t.Optional[t.Pattern]  # type: ignore[type-arg]

    @pydantic.validator("report_level", pre=True)
    @classmethod
    def valid_report_level(cls, value: t.Any) -> t.Optional[ReportLevel]:  # noqa: ANN401
        """Validate the report_level setting.

        :param value: Value to validate
        :raises ValueError: If ``Value`` is not a valid docutils report level
        :return: Instance of ``ReportLevel`` or None if emptry string.
        """
        if value is None:
            return None

        if isinstance(value, ReportLevel):
            return value

        if value == "":
            return DEFAULT_REPORT_LEVEL

        if isinstance(value, bool):
            raise ValueError("Invalid report level")

        if isinstance(value, str):
            if value.casefold() in set(ReportLevelMap):
                return ReportLevel(ReportLevelMap[value.casefold()])

            with contextlib.suppress(ValueError):
                value = int(value)

        if isinstance(value, int) and 1 <= value <= 5:
            return ReportLevel(value)

        raise ValueError("Invalid report level")

    @pydantic.validator(
        "ignore_directives", "ignore_roles", "ignore_substitutions", "ignore_languages", pre=True
    )
    @classmethod
    def split_str(cls, value: t.Any) -> t.Optional[t.List[str]]:  # noqa: ANN401
        """Validate and parse the following ignore_* settings.

        - ignore_directives
        - ignore_roles
        - ignore_substitutions
        - ignore_languages

        Comma separated strings are split into a list.

        :param value: Value to validate
        :raises ValueError: If not a ``str`` or ``List[str]``
        :return: List of things to ignore in the respective category
        """
        if value is None:
            return None

        if isinstance(value, str):
            return [v.strip() for v in value.split(",") if v.strip()]

        if isinstance(value, list) and all(isinstance(v, str) for v in value):
            return [v.strip() for v in value if v.strip()]

        raise ValueError("Not a string or list of strings")

    @pydantic.validator("ignore_messages", pre=True)
    @classmethod
    def join_regex_str(
        cls, value: t.Any  # noqa: ANN401
    ) -> t.Optional[t.Union[str, t.Pattern[str]]]:
        """Validate and concatenate the ignore_messages setting to a RegEx string.

        If a list ist given, the entries are concatenated with "|" to create an or RegEx.

        :param value: Value to validate
        :raises ValueError: If not a ``str`` or ``List[str]``
        :return: A RegEx string with messages to ignore or Pattern if it is one already
        """
        if value is None:
            return None

        if isinstance(value, re.Pattern):
            return value

        if isinstance(value, list) and all(isinstance(v, str) for v in value):
            return r"|".join(value)

        if isinstance(value, str):
            return value

        raise ValueError("Not a string or list of strings")


class RstcheckConfig(RstcheckConfigFile):  # pylint: disable=too-few-public-methods
    """Rstcheck config.

    :raises ValueError:: If setting has incorrect value or type
    :raises pydantic.error_wrappers.ValidationError:: If setting is not parsable into correct type
    """

    config_path: t.Optional[pathlib.Path]
    recursive: t.Optional[bool]


class _RstcheckConfigINIFile(
    pydantic.BaseModel  # pylint: disable=no-member
):  # pylint: disable=too-few-public-methods
    """Type for [rstcheck] section in INI file.

    The types apply to the file's data before the parsing by ``RstcheckConfig`` is done.

    :raises pydantic.error_wrappers.ValidationError:: If setting is not parsable into correct type
    """

    report_level: pydantic.NoneStr = pydantic.Field(None)  # pylint: disable=no-member
    ignore_directives: pydantic.NoneStr = pydantic.Field(None)  # pylint: disable=no-member
    ignore_roles: pydantic.NoneStr = pydantic.Field(None)  # pylint: disable=no-member
    ignore_substitutions: pydantic.NoneStr = pydantic.Field(None)  # pylint: disable=no-member
    ignore_languages: pydantic.NoneStr = pydantic.Field(None)  # pylint: disable=no-member
    ignore_messages: pydantic.NoneStr = pydantic.Field(None)  # pylint: disable=no-member


def _load_config_from_ini_file(ini_file: pathlib.Path) -> t.Optional[RstcheckConfigFile]:
    """Load, parse and validate rstcheck config from a ini file.

    :param ini_file: INI file to load config from
    :raises FileNotFoundError: If the file is not found
    :return: instance of ``RstcheckConfigFile`` or ``None`` on missing config section
    """
    resolved_file = ini_file.resolve()

    if not resolved_file.is_file():
        raise FileNotFoundError(f"{resolved_file}")

    parser = configparser.ConfigParser()
    parser.read(resolved_file)

    if not parser.has_section("rstcheck"):
        return None

    config_values_raw = dict(parser.items("rstcheck"))
    config_values_checked = _RstcheckConfigINIFile(**config_values_raw)
    config_values_parsed = RstcheckConfigFile(**config_values_checked.dict())

    return config_values_parsed


class _RstcheckConfigTOMLFile(
    pydantic.BaseModel  # pylint: disable=no-member,
):  # pylint: disable=too-few-public-methods
    """Type for [tool.rstcheck] section in TOML file.

    The types apply to the file's data before the parsing by ``RstcheckConfig`` is done.

    :raises pydantic.error_wrappers.ValidationError:: If setting is not parsable into correct type
    """

    report_level: t.Optional[str] = pydantic.Field(None)
    ignore_directives: t.Optional[t.List[str]] = pydantic.Field(None)
    ignore_roles: t.Optional[t.List[str]] = pydantic.Field(None)
    ignore_substitutions: t.Optional[t.List[str]] = pydantic.Field(None)
    ignore_languages: t.Optional[t.List[str]] = pydantic.Field(None)
    ignore_messages: t.Optional[t.Union[str, t.List[str]]] = pydantic.Field(None)


def _load_config_from_toml_file(toml_file: pathlib.Path) -> t.Optional[RstcheckConfigFile]:
    """Load, parse and validate rstcheck config from a TOML file.

    .. warning::

        Needs tomli installed!
        Use toml extra.

    :param toml_file: TOML file to load config from
    :raises ValueError: If the file is not a TOML file
    :raises FileNotFoundError: If the file is not found
    :return: instance of ``RstcheckConfigFile`` or ``None`` on missing config section
    """
    _extras.install_guard("tomli")

    resolved_file = toml_file.resolve()

    if resolved_file.suffix.casefold() != ".toml":
        raise ValueError("File is not a TOML file")

    if not resolved_file.is_file():
        raise FileNotFoundError(f"{resolved_file}")

    with open(resolved_file, "rb") as toml_file_handle:
        toml_dict = tomli.load(toml_file_handle)

    optional_rstcheck_section = t.Optional[t.Dict[str, t.Any]]
    rstcheck_section: optional_rstcheck_section = toml_dict.get("tool", {}).get("rstcheck")

    if rstcheck_section is None:
        return None

    config_values_checked = _RstcheckConfigTOMLFile(**rstcheck_section)
    config_values_parsed = RstcheckConfigFile(**config_values_checked.dict())

    return config_values_parsed


def load_config_file(file_path: pathlib.Path) -> t.Optional[RstcheckConfigFile]:
    """Load, parse and validate rstcheck config from a file.

    .. caution::

        If a TOML file is passed this function need tomli installed!
        Use toml extra or install manally.

    :param file_path: File to load config from
    :raises FileNotFoundError: If the file is not found
    :return: instance of ``RstcheckConfigFile`` or ``None`` on missing config section
    """
    if file_path.suffix.casefold() == ".toml":
        return _load_config_from_toml_file(file_path)
    return _load_config_from_ini_file(file_path)


def load_config_file_from_dir(dir_path: pathlib.Path) -> t.Optional[RstcheckConfigFile]:
    """Search, load, parse and validate rstcheck config from a directory.

    Searches files from ``CONFIG_FILES`` in the directory. If a file is found, try to load the
    config from it. If is has no config, search further.

    :param dir_path: Directory to search
    :return: instance of ``RstcheckConfigFile`` or
        ``None`` if no file is found or no file has a rstcheck section
    """
    config = None

    for file_name in CONFIG_FILES:
        file_path = (dir_path / file_name).resolve()
        if file_path.is_file():
            config = load_config_file(file_path)
            if config is not None:
                break

    return config


def load_config_file_from_dir_tree(dir_path: pathlib.Path) -> t.Optional[RstcheckConfigFile]:
    """Search, load, parse and validate rstcheck config from a directory tree.

    Searches files from ``CONFIG_FILES`` in the directory. If a file is found, try to load the
    config from it. If is has no config, search further. If no config is found in the directory
    search its parents one by one.

    :param dir_path: Directory to search
    :return: instance of ``RstcheckConfigFile`` or
        ``None`` if no file is found or no file has a rstcheck section
    """
    config = None

    search_dir = dir_path.resolve()

    while True:
        config = load_config_file_from_dir(search_dir)

        if config is not None:
            break

        parent_dir = search_dir.parent.resolve()
        if parent_dir == search_dir:
            break
        search_dir = parent_dir

    return config


def load_config_file_from_path(
    path: pathlib.Path, *, search_dir_tree: bool = False
) -> t.Optional[RstcheckConfigFile]:
    """Analyse the path and call the correct config file loader.

    :param path: Path to load config file from; can be a file or directory
    :param search_dir_tree: If the directory tree should be searched;
        only applies if ``path`` is a directory;
        defaults to False
    :return: instance of ``RstcheckConfigFile`` or
        ``None`` if no file is found or no file has a rstcheck section
    """
    resolved_path = path.resolve()

    if resolved_path.is_file():
        return load_config_file(resolved_path)

    if resolved_path.is_dir():
        if search_dir_tree:
            return load_config_file_from_dir_tree(resolved_path)
        return load_config_file_from_dir(resolved_path)

    return None


def merge_configs(
    config_base: RstcheckConfig,
    config_add: t.Union[RstcheckConfig, RstcheckConfigFile],
    *,
    config_add_is_dominant: bool = True,
) -> RstcheckConfig:
    """Merge two configs into a new one.

    :param config_base: The base config to merge into
    :param config_add: The config that is merged into the ``config_base``
    :param config_add_is_dominant: If the ``config_add`` overwrites values of ``config_base``;
        defaults to True
    :return: New merged config
    """
    sub_config: t.Union[RstcheckConfig, RstcheckConfigFile] = config_base
    sub_config_dict = sub_config.dict()
    for setting in dict(sub_config_dict):
        if sub_config_dict[setting] is None:
            del sub_config_dict[setting]

    dom_config: t.Union[RstcheckConfig, RstcheckConfigFile] = config_add
    dom_config_dict = dom_config.dict()
    for setting in dict(dom_config_dict):
        if dom_config_dict[setting] is None:
            del dom_config_dict[setting]

    if config_add_is_dominant is False:
        sub_config_dict, dom_config_dict = dom_config_dict, sub_config_dict

    merged_config_dict = {**sub_config_dict, **dom_config_dict}

    return RstcheckConfig(**merged_config_dict)



"""Tests for ``_docutils`` module."""
# pylint: disable=protected-access
import typing as t

import docutils.parsers.rst.directives as docutils_directives
import docutils.parsers.rst.roles as docutils_roles
import pytest

from rstcheck import _docutils, _extras


class TestIgnoreDirectivesAndRoles:
    """Test ``ignore_directives_and_roles`` function."""

    @staticmethod
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_with_empty_lists() -> None:
        """Test with empty lists."""
        directives: t.List[str] = []
        roles: t.List[str] = []

        _docutils.ignore_directives_and_roles(directives, roles)  # act

        assert not docutils_directives._directives
        assert not docutils_roles._roles

    @staticmethod
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_with_only_roles() -> None:
        """Test with only roles to add."""
        directives: t.List[str] = []
        roles = ["test_role"]

        _docutils.ignore_directives_and_roles(directives, roles)  # act

        assert not docutils_directives._directives
        assert "test_role" in docutils_roles._roles

    @staticmethod
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_with_only_directives() -> None:
        """Test with only directives to add."""
        directives = ["test_directive"]
        roles: t.List[str] = []

        _docutils.ignore_directives_and_roles(directives, roles)  # act

        assert "test_directive" in docutils_directives._directives
        assert not docutils_roles._roles

    @staticmethod
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_with_both() -> None:
        """Test with both."""
        directives = ["test_directive"]
        roles = ["test_role"]

        _docutils.ignore_directives_and_roles(directives, roles)  # act

        assert "test_directive" in docutils_directives._directives
        assert "test_role" in docutils_roles._roles


class TestRegisterCodeRirective:
    """Test ``register_code_directive`` function."""

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_does_nothing_when_sphinx_installed() -> None:
        """Test function does nothing when sphinx is installed."""
        _docutils.register_code_directive()  # act

        assert "code" not in docutils_directives._directives
        assert "code-block" not in docutils_directives._directives
        assert "sourcecode" not in docutils_directives._directives

    @staticmethod
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_registers_all_when_sphinx_is_missing() -> None:
        """Test function registers all directives when sphinx is missing."""
        _docutils.register_code_directive()  # act

        assert "code" in docutils_directives._directives
        assert "code-block" in docutils_directives._directives
        assert "sourcecode" in docutils_directives._directives

    @staticmethod
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_does_nothing_when_sphinx_is_missing_and_all_ignored() -> None:
        """Test function does nothing when sphinx is missing, but all ignores are ``True``."""
        _docutils.register_code_directive(  # act
            ignore_code_directive=True,
            ignore_codeblock_directive=True,
            ignore_sourcecode_directive=True,
        )

        assert "code" not in docutils_directives._directives
        assert "code-block" not in docutils_directives._directives
        assert "sourcecode" not in docutils_directives._directives

    @staticmethod
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_install_only_code_when_others_are_ignored() -> None:
        """Test function installes only code directive when others are ignored."""
        _docutils.register_code_directive(  # act
            ignore_codeblock_directive=True, ignore_sourcecode_directive=True
        )

        assert "code" in docutils_directives._directives
        assert "code-block" not in docutils_directives._directives
        assert "sourcecode" not in docutils_directives._directives

    @staticmethod
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_install_only_code_block_when_others_are_ignored() -> None:
        """Test function installes only code-block directive when others are ignored."""
        _docutils.register_code_directive(  # act
            ignore_code_directive=True, ignore_sourcecode_directive=True
        )

        assert "code" not in docutils_directives._directives
        assert "code-block" in docutils_directives._directives
        assert "sourcecode" not in docutils_directives._directives

    @staticmethod
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_install_only_sourcecode_when_others_are_ignored() -> None:
        """Test function installes only sourcecode directive when others are ignored."""
        _docutils.register_code_directive(  # act
            ignore_code_directive=True, ignore_codeblock_directive=True
        )

        assert "code" not in docutils_directives._directives
        assert "code-block" not in docutils_directives._directives
        assert "sourcecode" in docutils_directives._directives



"""Tests for ``_extras`` module."""
import pytest

from rstcheck import _compat, _extras


class TestInstallChecker:
    """Test ``is_installed_with_supported_version``."""

    @staticmethod
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    def test_false_on_missing_sphinx_package() -> None:
        """Test install-checker returns ``False`` when ``sphinx`` is missing."""
        result = _extras.is_installed_with_supported_version("sphinx")

        assert result is False

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    def test_true_on_installed_sphinx_package() -> None:
        """Test install-checker returns ``True`` when ``sphinx`` is installed with good version."""
        result = _extras.is_installed_with_supported_version("sphinx")

        assert result is True

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    def test_false_on_installed_sphinx_package_too_old(monkeypatch: pytest.MonkeyPatch) -> None:
        """Test install-checker returns ``False`` when ``sphinx`` is installed with bad version."""
        monkeypatch.setattr(_compat, "version", lambda _: "0.0")

        result = _extras.is_installed_with_supported_version("sphinx")

        assert result is False


class TestInstallGuard:
    """Test ``install_guard``."""

    @staticmethod
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    def test_false_on_missing_sphinx_package() -> None:
        """Test install-guard raises exception when ``sphinx`` is missing."""
        with pytest.raises(ModuleNotFoundError):
            _extras.install_guard("sphinx")  # act

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    def test_true_on_installed_sphinx_package() -> None:
        """Test install-guard doesn't raise when ``sphinx`` is installed."""
        _extras.install_guard("sphinx")  # act



# pylint: disable=too-many-lines
"""Tests for ``checker`` module."""
import os
import pathlib
import re
import shlex
import sys
from inspect import isfunction

import docutils.io
import docutils.nodes
import docutils.utils
import pytest

from rstcheck import _extras, checker, config, types


def test_check_file(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test ``check_file`` returns accumulated errors from ``check_source``."""
    errors = [
        types.LintError(source_origin="<string>", line_number=0, message=""),
        types.LintError(source_origin="<string>", line_number=1, message=""),
    ]
    monkeypatch.setattr(checker, "_get_source", lambda _: "source")
    monkeypatch.setattr(
        checker, "check_source", lambda _, source_file, ignores, report_level: (e for e in errors)
    )
    test_config = config.RstcheckConfig(config_path=pathlib.Path())

    result = checker.check_file(pathlib.Path(), test_config)

    assert result == errors


class TestRunConfigLoader:
    """Test ``_load_run_config`` function."""

    @staticmethod
    def test_global_config_path_set() -> None:
        """Test config path is set in main config.

        This results in no change -> return of main config.
        """
        test_config = config.RstcheckConfig(config_path=pathlib.Path())

        result = checker._load_run_config(  # pylint: disable=protected-access
            pathlib.Path(), test_config
        )

        assert result == test_config

    @staticmethod
    def test_no_config_file_in_dir_tree(monkeypatch: pytest.MonkeyPatch) -> None:
        """Test config path is unset in main config and file dir tree has no config file.

        This results in no change -> return of main config.
        """
        monkeypatch.setattr(config, "load_config_file_from_dir_tree", lambda _: None)
        test_config = config.RstcheckConfig()

        result = checker._load_run_config(  # pylint: disable=protected-access
            pathlib.Path(), test_config
        )

        assert result == test_config

    @staticmethod
    def test_config_file_in_dir_tree(monkeypatch: pytest.MonkeyPatch) -> None:
        """Test config path is unset in main config and file dir tree has a config file.

        This results in merge of configs -> return merged config.
        """
        test_file_config = config.RstcheckConfigFile(report_level=config.ReportLevel.SEVERE)
        monkeypatch.setattr(config, "load_config_file_from_dir_tree", lambda _: test_file_config)
        test_config = config.RstcheckConfig()

        result = checker._load_run_config(  # pylint: disable=protected-access
            pathlib.Path(), test_config, True
        )

        assert result is not None
        assert test_config != config.ReportLevel.SEVERE
        assert result.report_level == config.ReportLevel.SEVERE


class TestSourceGetter:
    """Test ``_get_source`` function."""

    @staticmethod
    def test_file_name_is_dash(monkeypatch: pytest.MonkeyPatch) -> None:
        """Test when file name is a dash, stdin is read."""
        source = "Teststring"
        monkeypatch.setattr(sys.stdin, "read", lambda: source)
        test_file = pathlib.Path("-")

        result = checker._get_source(test_file)  # pylint: disable=protected-access

        assert result == source

    @staticmethod
    def test_file_name_is_not_dash(tmp_path: pathlib.Path) -> None:
        """Test when file name is not a dash, the file is read."""
        source = """
Test
====
"""
        test_file = tmp_path / "testfile.rst"
        test_file.write_text(source)

        result = checker._get_source(test_file)  # pylint: disable=protected-access

        assert result == source


def test__replace_ignored_substitutions() -> None:
    """Test ``_replace_ignored_substitutions`` fucntion replaces substitutions."""
    source = "|Substitution1| |Substitution2|"

    result = checker._replace_ignored_substitutions(  # pylint: disable=protected-access
        source, ["Substitution1"]
    )

    assert result == "xSubstitution1x |Substitution2|"


def test__create_ignore_dict_from_config() -> None:
    """Test ``_create_ignore_dict_from_config`` fucntion creates ignore dict."""
    ignore_messages = r"foo/bar"
    ignore_messages_re = re.compile(ignore_messages)
    ignore_languages = ["python", "cpp"]
    ignore_directives = ["code"]
    ignore_roles = ["role"]
    ignore_substitutions = ["substi"]
    test_config = config.RstcheckConfig(
        ignore_messages=ignore_messages,
        ignore_languages=ignore_languages,
        ignore_directives=ignore_directives,
        ignore_roles=ignore_roles,
        ignore_substitutions=ignore_substitutions,
    )

    result = checker._create_ignore_dict_from_config(  # pylint: disable=protected-access
        test_config,
    )

    assert result == types.IgnoreDict(
        messages=ignore_messages_re,
        languages=ignore_languages,
        directives=ignore_directives,
        roles=ignore_roles,
        substitutions=ignore_substitutions,
    )


class TestSourceChecker:
    """Test ``check_source`` fucntion."""

    @staticmethod
    def test_empty_source() -> None:
        """Test empty source generates no errors."""
        source = ""

        result = list(checker.check_source(source))

        assert not result

    @staticmethod
    def test_lint_error_no_source_file() -> None:
        """Test lint error holds "<string>" if no source file is passed."""
        source = """
Test
===
"""

        result = list(checker.check_source(source))

        assert result[0]["source_origin"] == "<string>"

    @staticmethod
    def test_lint_error_with_source_file() -> None:
        """Test lint error holds source file."""
        test_file = pathlib.Path("test_file.rst")
        source = """
Test
===
"""

        result = list(checker.check_source(source, test_file))

        assert result[0]["source_origin"] == test_file

    @staticmethod
    def test_lint_error_returned_on_default_ignore() -> None:
        """Test lint error is returned with default ignores."""
        source = """
Test
===
"""

        result = list(checker.check_source(source))

        assert "Possible title underline, too short for the title" in result[0]["message"]

    @staticmethod
    def test_lint_error_skipped_on_set_ignore() -> None:
        """Test lint error is skipped with set ignores."""
        source = """
Test
===
"""
        ignores = types.IgnoreDict(
            messages=re.compile(r"Possible title underline, too short for the title"),
            languages=[],
            directives=[],
            roles=[],
            substitutions=[],
        )

        result = list(checker.check_source(source, ignores=ignores))

        assert not result

    @staticmethod
    def test_invalid_inline_ignore_config() -> None:
        """Test error on invalid inline ignore-languages is passed on."""
        source = """
.. rstcheck: ignore-languages python
"""

        result = list(checker.check_source(source))

        assert 'Expected "key=value" syntax' in result[0].get("message", "")

    @staticmethod
    def test_invalid_inline_ignore_config_not_ignored() -> None:
        """Test error on invalid inline ignore-languages is not ignored."""
        source = """
.. rstcheck: ignore-languages python
"""
        ignores = types.IgnoreDict(
            messages=re.compile(r'Expected "key=value" syntax'),
            languages=[],
            directives=[],
            roles=[],
            substitutions=[],
        )

        result = list(checker.check_source(source, ignores=ignores))

        assert 'Expected "key=value" syntax' in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    def test_sphinx_directive_errors_without_sphinx() -> None:
        """Test error on sphinx directive when sphinx is missing."""
        source = """
.. py:function:: foo
"""

        result = list(checker.check_source(source))

        assert 'No directive entry for "py:function"' in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    def test_sphinx_directive_does_not_error_with_sphinx() -> None:
        """Test no error on sphinx directive when sphinx is installed."""
        source = """
.. py:function:: foo
"""

        result = list(checker.check_source(source))

        assert not result

    @staticmethod
    @pytest.mark.skipif(sys.version_info[0:2] > (3, 9), reason="Requires python3.9 or lower")
    def test_code_block_lint_error_returned_on_default_ignore_pre310() -> None:
        """Test code lint error is returned with default ignores.

        In Python version 3.10 the error messag changed.
        """
        source = """
.. code:: python

    print(
"""

        result = list(checker.check_source(source))

        assert "unexpected EOF while parsing" in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(sys.version_info[0:2] > (3, 9), reason="Requires python3.9 or lower")
    def test_code_block_no_error_on_set_ignore_pre310() -> None:
        """Test code lint error is skipped with set ignores.

        In Python version 3.10 the error messag changed.
        """
        source = """
.. code:: python

    print(
"""
        ignores = types.IgnoreDict(
            messages=re.compile(r"unexpected EOF while parsing"),
            languages=[],
            directives=[],
            roles=[],
            substitutions=[],
        )

        result = list(checker.check_source(source, ignores=ignores))

        assert not result

    @staticmethod
    @pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires python3.10 or higher")
    def test_code_block_lint_error_returned_on_default_ignore() -> None:
        """Test code lint error is returned with default ignores."""
        source = """
.. code:: python

    print(
"""

        result = list(checker.check_source(source))

        assert "'(' was never closed" in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires python3.10 or higher")
    def test_code_block_no_error_on_set_ignore() -> None:
        """Test code lint error is skipped with set ignores."""
        source = """
.. code:: python

    print(
"""
        ignores = types.IgnoreDict(
            messages=re.compile(r"'\(' was never closed"),
            languages=[],
            directives=[],
            roles=[],
            substitutions=[],
        )

        result = list(checker.check_source(source, ignores=ignores))

        assert not result


class TestCodeCheckRunner:
    """Test ``_run_code_checker_and_filter_errors`` function."""

    @staticmethod
    def test_without_ignore() -> None:
        """Test both checkers return error without ignore."""
        cb_checker = checker.CodeBlockChecker("<string>")
        checker_list = [
            cb_checker.create_checker("print(", "python"),
            cb_checker.create_checker("{", "json"),
        ]

        result = list(
            checker._run_code_checker_and_filter_errors(  # pylint: disable=protected-access
                checker_list, None
            )
        )

        assert len(result) == 2

    @staticmethod
    def test_with_ignore() -> None:
        """Test only one checker return error with ignore."""
        cb_checker = checker.CodeBlockChecker("<string>")
        checker_list = [
            cb_checker.create_checker("print(", "python"),
            cb_checker.create_checker("{", "json"),
        ]
        ignore_messages = re.compile(r"Expecting property name enclosed in double quotes")

        result = list(
            checker._run_code_checker_and_filter_errors(  # pylint: disable=protected-access
                checker_list, ignore_messages
            )
        )

        assert len(result) == 1


class TestRstErrorParseFilter:
    """Test ``_parse_and_filter_rst_errors`` function."""

    @staticmethod
    def test_without_ignore() -> None:
        """Test both error messages are parsed and returned."""
        error_str = "<string>:1:1: Error message 1\n<string>:1:2: Error message 2"

        result = list(
            checker._parse_and_filter_rst_errors(  # pylint: disable=protected-access
                error_str, "<string>", None
            )
        )

        assert len(result) == 2

    @staticmethod
    def test_with_ignore() -> None:
        """Test only one error message is parsed and returned."""
        error_str = "<string>:1:1: Error message 1\n<string>:1:2: Error message 2"
        ignore_messages = re.compile(r"Error message 1")

        result = list(
            checker._parse_and_filter_rst_errors(  # pylint: disable=protected-access
                error_str, "<string>", ignore_messages
            )
        )

        assert len(result) == 1


class TestCheckTranslator:  # pylint: disable=too-few-public-methods
    """Test ``_CheckTranslator`` class."""

    @staticmethod
    def test_no_checkers_on_init() -> None:
        """Test checkers are empty on init."""
        doc = docutils.utils.new_document("")

        result = checker._CheckTranslator(doc, "", "<string>")  # pylint: disable=protected-access

        assert not result.checkers


class TestCodeBlockChecker:  # pylint: disable=too-many-public-methods
    """Test ``CodeBlockChecker`` class."""

    @staticmethod
    def test_init() -> None:
        """Test nothing special happens on ``__init__`` method."""
        source_origin: types.SourceFileOrString = "<string>"

        result = checker.CodeBlockChecker(source_origin)

        assert result.source_origin == source_origin
        assert result.ignores is None
        assert result.report_level == config.ReportLevel.INFO

    @staticmethod
    def test_language_is_supported_on_supported_lang() -> None:
        """Test ``language_is_supported`` method returns ``True`` for supported language."""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = cb_checker.language_is_supported("python")

        assert result is True

    @staticmethod
    def test_language_is_supported_on_unsupported_lang() -> None:
        """Test ``language_is_supported`` method returns ``False`` for unsupported language."""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = cb_checker.language_is_supported("some-unsupported-lang")

        assert result is False

    @staticmethod
    def test_create_checker_returns_function() -> None:
        """Test ``create_checker`` method returns a lambda function.

        Currently unknown how to test if lambda body is ``self.check``.
        """
        cb_checker = checker.CodeBlockChecker("<string>")

        result = cb_checker.create_checker("", "")

        assert isfunction(result)
        assert result.__name__ == "<lambda>"

    @staticmethod
    def test_check_returns_none_on_unsupported_lang() -> None:
        """Test ``check`` returns ``None`` on unsupported language."""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check("", "some-unsupported-lang"))

        assert not result

    @staticmethod
    def test_check_returns_none_on_ok_code_block_for_supported_lang() -> None:
        """Test ``check`` returns ``None`` on ok code block for supported language."""
        source = """
print("rstcheck")
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check(source, "python"))

        assert not result

    @staticmethod
    @pytest.mark.skipif(sys.version_info[0:2] > (3, 9), reason="Requires python3.9 or lower")
    def test_check_returns_error_on_bad_code_block_for_supported_lang_pre310() -> None:
        """Test ``check`` returns error on bad code block for supported language.

        In Python version 3.10 the error messag changed.
        """
        source = """
print(
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check(source, "python"))

        assert "unexpected EOF while parsing" in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires python3.10 or higher")
    def test_check_returns_error_on_bad_code_block_for_supported_lang() -> None:
        """Test ``check`` returns error on bad code block for supported language."""
        source = """
print(
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check(source, "python"))

        assert "'(' was never closed" in result[0]["message"]

    @staticmethod
    def test_check_python_returns_none_on_ok_code_block() -> None:
        """Test ``check_python`` returns ``None`` on ok code block."""
        source = """
print("rstcheck")
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_python(source))

        assert not result

    @staticmethod
    @pytest.mark.skipif(sys.version_info[0:2] > (3, 9), reason="Requires python3.9 or lower")
    def test_check_python_returns_error_on_bad_code_block_pre310() -> None:
        """Test ``check_python`` returns error on bad code block.

        In Python version 3.10 the error messag changed.
        """
        source = """
print(
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_python(source))

        assert "unexpected EOF while parsing" in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires python3.10 or higher")
    def test_check_python_returns_error_on_bad_code_block() -> None:
        """Test ``check_python`` returns error on bad code block."""
        source = """
print(
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_python(source))

        assert "'(' was never closed" in result[0]["message"]

    @staticmethod
    def test_check_json_returns_none_on_ok_code_block() -> None:
        """Test ``check_json`` returns ``None`` on ok code block."""
        source = """
{
    "key": "value"
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_json(source))

        assert not result

    @staticmethod
    def test_check_json_returns_error_on_bad_code_block() -> None:
        """Test ``check_json`` returns error on bad code block."""
        source = """
{
    "key":
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_json(source))

        assert "Expecting value:" in result[0]["message"]

    @staticmethod
    def test_check_xml_returns_none_on_ok_code_block() -> None:
        """Test ``check_xml`` returns ``None`` on ok code block."""
        source = """<?xml version="1.0" encoding="UTF-8"?>
<note>
    <heading>Reminder</heading>
</note>
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_xml(source))

        assert not result

    @staticmethod
    def test_check_xml_returns_error_on_bad_code_block() -> None:
        """Test ``check_xml`` returns error on bad code block."""
        source = """<?xml version="1.0" encoding="UTF-8"?>
<note>
    <heading>Reminder
</note>
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_xml(source))

        assert "mismatched tag:" in result[0]["message"]

    @staticmethod
    def test_check_rst_returns_none_on_ok_code_block() -> None:
        """Test ``check_rst`` returns ``None`` on ok code block."""
        source = """
Heading
=======
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_rst(source))

        assert not result

    @staticmethod
    def test_check_rst_returns_error_on_bad_code_block() -> None:
        """Test ``check_rst`` returns error on bad code block."""
        source = """
Heading
======
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_rst(source))

        assert "Title underline too short." in result[0]["message"]

    @staticmethod
    def test_check_doctest_returns_none_on_ok_code_block() -> None:
        """Test ``check_doctest`` returns ``None`` on ok code block."""
        source = """
>>> x = 1
>>> x
1
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_doctest(source))

        assert not result

    @staticmethod
    def test_check_doctest_returns_error_on_bad_code_block() -> None:
        """Test ``check_doctest`` returns error on bad code block."""
        source = """
>>> x = 1
>>>> x
1
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_doctest(source))

        assert "lacks blank after >>>: '>>>> x'" in result[0]["message"]

    @staticmethod
    def test_check_doctest_returns_none_on_bad_code_block_if_regex_no_matching(
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test ``check_doctest`` returns ``None`` on bad code block if regex is not matching."""
        monkeypatch.setattr(checker, "DOCTEST_LINE_NO_REGEX", re.compile("bad-regex"))
        source = """
>>> x = 1
>>>> x
1
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_doctest(source))

        assert not result

    @staticmethod
    def test_check_bash_returns_none_on_ok_code_block() -> None:
        """Test ``check_bash`` returns ``None`` on ok code block."""
        source = """
if [ "$x" == 'y' ]
then
    exit 1
fi
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_bash(source))

        assert not result

    @staticmethod
    @pytest.mark.skipif(
        sys.platform == "win32",
        reason="Unknown Windows specific wrong positiv. `list index out of range`",
    )
    def test_check_bash_returns_error_on_bad_code_block() -> None:
        """Test ``check_bash`` returns error on bad code block."""
        source = """
{
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_bash(source))

        assert "syntax error: unexpected end of file" in result[0]["message"]

    @staticmethod
    def test_check_c_returns_none_on_ok_code_block() -> None:
        """Test ``check_c`` returns ``None`` on ok code block."""
        source = """
float foo(int n)
{
    // Test C99.
    float x[n];
    x[0] = 1;
    return x[0];
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_c(source))

        assert not result

    @staticmethod
    @pytest.mark.skipif(sys.platform != "linux", reason="Linux specific error message")
    def test_check_c_returns_error_on_bad_code_block_linux() -> None:
        """Test ``check_c`` returns error on bad code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_c(source))

        assert (
            "error: \u2018x\u2019 undeclared (first use in this function)" in result[0]["message"]
        )

    @staticmethod
    @pytest.mark.skipif(sys.platform != "darwin", reason="MacOS specific error message")
    def test_check_c_returns_error_on_bad_code_block_macos() -> None:
        """Test ``check_c`` returns error on bad code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_c(source))

        assert "error: use of undeclared identifier 'x'" in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(sys.platform != "win32", reason="Windows specific error message")
    def test_check_c_returns_error_on_bad_code_block_windows() -> None:
        """Test ``check_c`` returns error on bad code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_c(source))

        assert "error: 'x' undeclared (first use in this function)" in result[0]["message"]

    @staticmethod
    def test_check_cpp_returns_none_on_ok_code_block() -> None:
        """Test ``check_cpp`` returns ``None`` on ok code block."""
        source = """
#include <iostream>

int main()
{
    auto x = 1;
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_cpp(source))

        assert not result

    @staticmethod
    @pytest.mark.skipif(sys.platform != "linux", reason="Linux specific error message")
    def test_check_cpp_returns_error_on_bad_code_block_linux() -> None:
        """Test ``check_cpp`` returns error on bad code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_cpp(source))

        assert "error: \u2018x\u2019 was not declared in this scope" in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(sys.platform != "darwin", reason="MacOS specific error message")
    def test_check_cpp_returns_error_on_bad_code_block_macos() -> None:
        """Test ``check_cpp`` returns error on bad code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_cpp(source))

        assert "error: use of undeclared identifier 'x'" in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(sys.platform != "win32", reason="Windows specific error message")
    def test_check_cpp_returns_error_on_bad_code_block_windows() -> None:
        """Test ``check_cpp`` returns error on bad code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(cb_checker.check_cpp(source))

        assert "error: 'x' was not declared in this scope" in result[0]["message"]

    @staticmethod
    def test__gcc_checker_returns_none_on_ok_cpp_code_block() -> None:
        """Test ``_gcc_checker`` returns ``None`` on ok c++ code block."""
        source = """
#include <iostream>

int main()
{
    auto x = 1;
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(
            cb_checker._gcc_checker(  # pylint: disable=protected-access
                source,
                ".cpp",
                [os.getenv("CXX", "g++")]
                + shlex.split(os.getenv("CXXFLAGS", ""))
                + shlex.split(os.getenv("CPPFLAGS", ""))
                + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
            )
        )

        assert not result

    @staticmethod
    @pytest.mark.skipif(sys.platform != "linux", reason="Linux specific error message")
    def test__gcc_checker_returns_error_on_bad_cpp_code_block_linux() -> None:
        """Test ``_gcc_checker`` returns error on bad c++ code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(
            cb_checker._gcc_checker(  # pylint: disable=protected-access
                source,
                ".cpp",
                [os.getenv("CXX", "g++")]
                + shlex.split(os.getenv("CXXFLAGS", ""))
                + shlex.split(os.getenv("CPPFLAGS", ""))
                + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
            )
        )

        assert "error: \u2018x\u2019 was not declared in this scope" in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(sys.platform != "darwin", reason="MacOS specific error message")
    def test__gcc_checker_returns_error_on_bad_cpp_code_block_macos() -> None:
        """Test ``_gcc_checker`` returns error on bad c++ code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(
            cb_checker._gcc_checker(  # pylint: disable=protected-access
                source,
                ".cpp",
                [os.getenv("CXX", "g++")]
                + shlex.split(os.getenv("CXXFLAGS", ""))
                + shlex.split(os.getenv("CPPFLAGS", ""))
                + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
            )
        )

        assert "error: use of undeclared identifier 'x'" in result[0]["message"]

    @staticmethod
    @pytest.mark.skipif(sys.platform != "win32", reason="Windows specific error message")
    def test__gcc_checker_returns_error_on_bad_cpp_code_block_windows() -> None:
        """Test ``_gcc_checker`` returns error on bad c++ code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = list(
            cb_checker._gcc_checker(  # pylint: disable=protected-access
                source,
                ".cpp",
                [os.getenv("CXX", "g++")]
                + shlex.split(os.getenv("CXXFLAGS", ""))
                + shlex.split(os.getenv("CPPFLAGS", ""))
                + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
            )
        )

        assert "error: 'x' was not declared in this scope" in result[0]["message"]

    @staticmethod
    def test__run_in_subprocess_returns_none_on_ok_cpp_code_block() -> None:
        """Test ``_run_in_subprocess`` returns ``None`` on ok c++ code block."""
        source = """
#include <iostream>

int main()
{
    auto x = 1;
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = cb_checker._run_in_subprocess(  # pylint: disable=protected-access
            source,
            ".cpp",
            [os.getenv("CXX", "g++")]
            + shlex.split(os.getenv("CXXFLAGS", ""))
            + shlex.split(os.getenv("CPPFLAGS", ""))
            + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
        )

        assert not result

    @staticmethod
    @pytest.mark.skipif(sys.platform != "linux", reason="Linux specific error message")
    def test__run_in_subprocess_returns_error_on_bad_cpp_code_block_linux() -> None:
        """Test ``_run_in_subprocess`` returns error on bad c++ code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = cb_checker._run_in_subprocess(  # pylint: disable=protected-access
            source,
            ".cpp",
            [os.getenv("CXX", "g++")]
            + shlex.split(os.getenv("CXXFLAGS", ""))
            + shlex.split(os.getenv("CPPFLAGS", ""))
            + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
        )

        assert result is not None
        assert "error: \u2018x\u2019 was not declared in this scope" in result[0]
        assert result[1].suffix == ".cpp"

    @staticmethod
    @pytest.mark.skipif(sys.platform != "darwin", reason="MacOS specific error message")
    def test__run_in_subprocess_returns_error_on_bad_cpp_code_block_macos() -> None:
        """Test ``_run_in_subprocess`` returns error on bad c++ code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = cb_checker._run_in_subprocess(  # pylint: disable=protected-access
            source,
            ".cpp",
            [os.getenv("CXX", "g++")]
            + shlex.split(os.getenv("CXXFLAGS", ""))
            + shlex.split(os.getenv("CPPFLAGS", ""))
            + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
        )

        assert result is not None
        assert "error: use of undeclared identifier 'x'" in result[0]
        assert result[1].suffix == ".cpp"

    @staticmethod
    @pytest.mark.skipif(sys.platform != "win32", reason="Windows specific error message")
    def test__run_in_subprocess_returns_error_on_bad_cpp_code_block_windows() -> None:
        """Test ``_run_in_subprocess`` returns error on bad c++ code block."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker("<string>")

        result = cb_checker._run_in_subprocess(  # pylint: disable=protected-access
            source,
            ".cpp",
            [os.getenv("CXX", "g++")]
            + shlex.split(os.getenv("CXXFLAGS", ""))
            + shlex.split(os.getenv("CPPFLAGS", ""))
            + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
        )

        assert result is not None
        assert "error: 'x' was not declared in this scope" in result[0]
        assert result[1].suffix == ".cpp"

    @staticmethod
    @pytest.mark.skipif(sys.platform != "linux", reason="Linux specific error message")
    def test__run_in_subprocess_returns_error_on_bad_cpp_code_block_with_filename_linux() -> None:
        """Test ``_run_in_subprocess`` returns error on bad c++ code block from filename."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker(pathlib.Path("filename.cpp"))

        result = cb_checker._run_in_subprocess(  # pylint: disable=protected-access
            source,
            ".cpp",
            [os.getenv("CXX", "g++")]
            + shlex.split(os.getenv("CXXFLAGS", ""))
            + shlex.split(os.getenv("CPPFLAGS", ""))
            + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
        )

        assert result is not None
        assert "error: \u2018x\u2019 was not declared in this scope" in result[0]
        assert result[1].suffix == ".cpp"

    @staticmethod
    @pytest.mark.skipif(sys.platform != "darwin", reason="MacOS specific error message")
    def test__run_in_subprocess_returns_error_on_bad_cpp_code_block_with_filename_macos() -> None:
        """Test ``_run_in_subprocess`` returns error on bad c++ code block from filename."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker(pathlib.Path("filename.cpp"))

        result = cb_checker._run_in_subprocess(  # pylint: disable=protected-access
            source,
            ".cpp",
            [os.getenv("CXX", "g++")]
            + shlex.split(os.getenv("CXXFLAGS", ""))
            + shlex.split(os.getenv("CPPFLAGS", ""))
            + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
        )

        assert result is not None
        assert "error: use of undeclared identifier 'x'" in result[0]
        assert result[1].suffix == ".cpp"

    @staticmethod
    @pytest.mark.skipif(sys.platform != "win32", reason="Windows specific error message")
    def test__run_in_subprocess_returns_error_on_bad_cpp_code_block_with_filename_windows() -> None:
        """Test ``_run_in_subprocess`` returns error on bad c++ code block from filename."""
        source = """
int main()
{
    return x;
}
"""
        cb_checker = checker.CodeBlockChecker(pathlib.Path("filename.cpp"))

        result = cb_checker._run_in_subprocess(  # pylint: disable=protected-access
            source,
            ".cpp",
            [os.getenv("CXX", "g++")]
            + shlex.split(os.getenv("CXXFLAGS", ""))
            + shlex.split(os.getenv("CPPFLAGS", ""))
            + ["-I.", "-I..", "-pedantic", "-fsyntax-only"],
        )

        assert result is not None
        assert "error: 'x' was not declared in this scope" in result[0]
        assert result[1].suffix == ".cpp"

    @staticmethod
    def test__parse_gcc_style_error_message_raises_on_bad_format() -> None:
        """Test ``_parse_gcc_style_error_message`` method raises ``ValueError`` on bad format."""
        message = "Foo bar"

        with pytest.raises(ValueError, match="Message cannot be parsed."):
            checker._parse_gcc_style_error_message(  # pylint: disable=protected-access
                message, "<string>"
            )

    @staticmethod
    def test__parse_gcc_style_error_message_with_column() -> None:
        """Test ``_parse_gcc_style_error_message`` method with column."""
        message = "<string>:16:32: Error message"
        error = types.LintError(
            source_origin="<string>",
            line_number=16,
            message="Error message",
        )

        result = checker._parse_gcc_style_error_message(  # pylint: disable=protected-access
            message, "<string>"
        )

        assert result == error

    @staticmethod
    def test__parse_gcc_style_error_message_without_column() -> None:
        """Test ``_parse_gcc_style_error_message`` method without column."""
        message = "<string>:16: Error message"
        error = types.LintError(
            source_origin="<string>",
            line_number=16,
            message="Error message",
        )

        result = checker._parse_gcc_style_error_message(  # pylint: disable=protected-access
            message, "<string>", False
        )

        assert result == error



"""Central place for install-checker and guards for 'extras' dependencies.

The ``*_INSTALLED`` constansts reveal wether the dependency is installed with a supported version.
The ``install_guard`` guard function is intended for use inside functions which need specific
extra packages installed.

Example usage:

.. code-block:: python

    from rstcheck import _extras

    if _extras.SPHINX_INSTALLED:
        import sphinx


    def print_sphinx_version():
        _extras.install_guard("sphinx")
        print(sphinx.version_info)
"""
import importlib
import typing as t

from . import _compat


ExtraDependencies = _compat.Literal["sphinx", "tomli"]
"""List of all dependencies installable through extras."""


class DependencyInfos(_compat.TypedDict):
    """Information about a dependency."""

    min_version: t.Tuple[int, ...]
    extra: str


ExtraDependenciesInfos: t.Dict[ExtraDependencies, DependencyInfos] = {
    "sphinx": DependencyInfos(min_version=(2, 0), extra="sphinx"),
    "tomli": DependencyInfos(min_version=(2, 0), extra="toml"),
}
"""Dependency map with their min. supported version and extra by which they can be installed."""


def is_installed_with_supported_version(package: ExtraDependencies) -> bool:
    """Check if the package is installed and has the minimum required version.

    :param package: Name of packge to check
    :return: Bool if package is installed with supported version
    """
    try:
        importlib.import_module(package)
    except ImportError:
        return False

    version: str = _compat.version(package)
    version_tuple = tuple(int(v) for v in version.split(".")[:3])

    return version_tuple >= ExtraDependenciesInfos[package]["min_version"]


SPHINX_INSTALLED = is_installed_with_supported_version("sphinx")
TOMLI_INSTALLED = is_installed_with_supported_version("tomli")


ExtraDependenciesInstalled: t.Dict[ExtraDependencies, bool] = {
    "sphinx": SPHINX_INSTALLED,
    "tomli": TOMLI_INSTALLED,
}


def install_guard(package: ExtraDependencies) -> None:
    """Guard code that needs the ``package`` installed and throw ModuleNotFoundError.

    See example in module docstring.

    :param package: Name of packge to check
    :raises ModuleNotFoundError: When the package is not installed.
    """
    if ExtraDependenciesInstalled[package] is True:
        return

    extra = ExtraDependenciesInfos[package]

    raise ModuleNotFoundError(
        f"No supported version of {package} installed. "
        f"Install rstcheck with {extra} extra (rstcheck[{extra}]) or "
        f"install a supported version of {package} yourself."
    )



"""Sphinx helper functions."""
import contextlib
import pathlib
import tempfile
import typing as t

from . import _docutils, _extras


if _extras.SPHINX_INSTALLED:
    import sphinx.application
    import sphinx.domains.c
    import sphinx.domains.cpp
    import sphinx.domains.javascript
    import sphinx.domains.python
    import sphinx.domains.std


@contextlib.contextmanager
def load_sphinx_if_available() -> t.Generator[None, None, None]:
    """Contextmanager to register Sphinx directives and roles if sphinx is available."""
    if _extras.SPHINX_INSTALLED:
        with tempfile.TemporaryDirectory() as temp_dir:
            outdir = pathlib.Path(temp_dir) / "_build"
            sphinx.application.Sphinx(
                srcdir=temp_dir,
                confdir=None,
                outdir=str(outdir),
                doctreedir=str(outdir),
                buildername="dummy",
                status=None,  # type: ignore[arg-type] # NOTE: sphinx type hint is incorrect
            )
            # NOTE: Hack to prevent sphinx warnings for overwriting registered nodes; see #113
            sphinx.application.builtin_extensions = [  # type: ignore[assignment]
                e for e in sphinx.application.builtin_extensions if e != "sphinx.addnodes"
            ]
            yield
    else:
        yield


def get_sphinx_directives_and_roles() -> t.Tuple[t.List[str], t.List[str]]:
    """Return Sphinx directives and roles loaded from sphinx.

    :return: Tuple of directives and roles
    """
    _extras.install_guard("sphinx")

    sphinx_directives = list(sphinx.domains.std.StandardDomain.directives)
    sphinx_roles = list(sphinx.domains.std.StandardDomain.roles)

    for domain in [
        sphinx.domains.c.CDomain,
        sphinx.domains.cpp.CPPDomain,
        sphinx.domains.javascript.JavaScriptDomain,
        sphinx.domains.python.PythonDomain,
    ]:
        domain_directives = list(domain.directives)
        domain_roles = list(domain.roles)

        sphinx_directives += domain_directives + [
            f"{domain.name}:{item}" for item in domain_directives
        ]

        sphinx_roles += domain_roles + [f"{domain.name}:{item}" for item in domain_roles]

    sphinx_directives += list(
        sphinx.application.docutils.directives._directives  # pylint: disable=protected-access
    )
    sphinx_roles += list(
        sphinx.application.docutils.roles._roles  # pylint: disable=protected-access
    )

    return (sphinx_directives, sphinx_roles)


_DIRECTIVE_WHITELIST = ["code", "code-block", "include"]
_ROLE_WHITELIST: t.List[str] = []


def filter_whitelisted_directives_and_roles(
    directives: t.List[str], roles: t.List[str]
) -> t.Tuple[t.List[str], t.List[str]]:
    """Filter whitelisted directives and roles out of input.

    :param directives: Directives to filter
    :param roles: Roles to filter
    :return: Tuple of fitlered directives and roles
    """
    directives = list(filter(lambda d: d not in _DIRECTIVE_WHITELIST, directives))
    roles = list(filter(lambda r: r not in _ROLE_WHITELIST, roles))

    return (directives, roles)


def load_sphinx_ignores() -> None:  # pragma: no cover
    """Register Sphinx directives and roles to ignore."""
    _extras.install_guard("sphinx")

    (directives, roles) = get_sphinx_directives_and_roles()
    (directives, roles) = filter_whitelisted_directives_and_roles(directives, roles)

    _docutils.ignore_directives_and_roles(directives, roles)



"""Tests for ``_sphinx`` module."""
# pylint: disable=protected-access
import typing as t

import docutils.parsers.rst.directives as docutils_directives
import docutils.parsers.rst.roles as docutils_roles
import pytest

from rstcheck import _extras, _sphinx


if _extras.SPHINX_INSTALLED:
    import sphinx.application


class TestContextManager:
    """Test ``load_sphinx_if_available`` context manager."""

    @staticmethod  # noqa: AAA01
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_yield_nothing_with_sphinx_missing() -> None:
        """Test for ``None`` yield and no action when sphinx is missing."""
        with _sphinx.load_sphinx_if_available() as ctx_manager:

            assert ctx_manager is None
            assert not docutils_directives._directives
            assert not docutils_roles._roles

    @staticmethod  # noqa: AAA01
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_yield_nothing_with_sphinx_installed() -> None:
        """Test for ``None`` yield but action when sphinx is installed."""
        with _sphinx.load_sphinx_if_available() as ctx_manager:

            assert ctx_manager is None
            assert docutils_directives._directives
            assert docutils_roles._roles
            assert "sphinx.addnodes" not in sphinx.application.builtin_extensions


class TestSphinxDirectiveAndRoleGetter:
    """Test ``get_sphinx_directives_and_roles`` function."""

    @staticmethod
    @pytest.mark.skipif(_extras.SPHINX_INSTALLED, reason="Test without sphinx extra.")
    def test_exception_on_missing_sphinx() -> None:
        """Test that the install guard triggers."""
        with pytest.raises(ModuleNotFoundError):
            _sphinx.get_sphinx_directives_and_roles()

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_c_domain_is_loaded() -> None:
        """Test C domain is loaded."""
        (result_directives, result_roles) = _sphinx.get_sphinx_directives_and_roles()  # act

        assert "function" in result_directives
        assert "c:function" in result_directives
        assert "member" in result_roles
        assert "c:member" in result_roles

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_cpp_domain_is_loaded() -> None:
        """Test C++ domain is loaded."""
        (result_directives, result_roles) = _sphinx.get_sphinx_directives_and_roles()  # act

        assert "function" in result_directives
        assert "cpp:function" in result_directives
        assert "member" in result_roles
        assert "cpp:member" in result_roles

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_javascript_domain_is_loaded() -> None:
        """Test JavaScript domain is loaded."""
        (result_directives, result_roles) = _sphinx.get_sphinx_directives_and_roles()  # act

        assert "function" in result_directives
        assert "js:function" in result_directives
        assert "func" in result_roles
        assert "js:func" in result_roles

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    @pytest.mark.usefixtures("patch_docutils_directives_and_roles_dict")
    def test_python_domain_is_loaded() -> None:
        """Test Python domain is loaded."""
        (result_directives, result_roles) = _sphinx.get_sphinx_directives_and_roles()  # act

        assert "function" in result_directives
        assert "py:function" in result_directives
        assert "func" in result_roles
        assert "py:func" in result_roles

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    def test_docutils_state_dict_is_loaded(monkeypatch: pytest.MonkeyPatch) -> None:
        """Test docutils' state is loaded."""
        test_dict_directives: t.Dict[str, t.Any] = {"test-directive": "test-directive"}
        monkeypatch.setattr(
            sphinx.application.docutils.directives, "_directives", test_dict_directives
        )
        test_dict_roles: t.Dict[str, t.Any] = {"test-role": "test-role"}
        monkeypatch.setattr(sphinx.application.docutils.roles, "_roles", test_dict_roles)

        (result_directives, result_roles) = _sphinx.get_sphinx_directives_and_roles()  # act

        assert "test-directive" in result_directives
        assert "test-role" in result_roles


class TestDirectiveAndRoleFilter:
    """Test ``filter_whitelisted_directives_and_roles`` function."""

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    def test_directives_are_filtered(monkeypatch: pytest.MonkeyPatch) -> None:
        """Test directives are filtered."""
        monkeypatch.setattr(_sphinx, "_DIRECTIVE_WHITELIST", ["test-directive"])
        unfiltered_directives = ["test-directive", "test-directive2"]

        (result_directives, _) = _sphinx.filter_whitelisted_directives_and_roles(
            unfiltered_directives, []
        )  # act

        assert "test-directive" not in result_directives
        assert "test-directive2" in result_directives

    @staticmethod
    @pytest.mark.skipif(not _extras.SPHINX_INSTALLED, reason="Depends on sphinx extra.")
    def test_roles_are_filtered(monkeypatch: pytest.MonkeyPatch) -> None:
        """Test roles are filtered."""
        monkeypatch.setattr(_sphinx, "_ROLE_WHITELIST", ["test-role"])
        unfiltered_roles = ["test-role", "test-role2"]

        (_, result_roles) = _sphinx.filter_whitelisted_directives_and_roles(
            [], unfiltered_roles
        )  # act

        assert "test-role" not in result_roles
        assert "test-role2" in result_roles



"""Inline config comment functionality."""
import re
import typing as t


RSTCHECK_COMMENT_REGEX = re.compile(r"\.\. rstcheck:")


class RstcheckCommentSyntaxError(Exception):
    """Syntax error for rstcheck inline config comments."""

    def __init__(self, message: str, line_number: int) -> None:
        """Initialize the ``RstcheckCommentSyntaxError`` exception.

        :param message: Error message
        :param line_number: Line number where the error occured
        """
        self.line_number = line_number
        Exception.__init__(self, message)


def find_ignored_languages(source: str) -> t.Generator[str, None, None]:  # noqa: CCR001
    """Search the rst source for rstcheck inline ignore-languages comments.

    Languages are ignored via comment.

    For example, to ignore C++, JSON, and Python:

    >>> list(find_ignored_languages('''
    ... Example
    ... =======
    ...
    ... .. rstcheck: ignore-languages=cpp,json
    ...
    ... .. rstcheck: ignore-languages=python
    ... '''))
    ["cpp", "json", "python"]

    :param source: Rst source code
    :raises RstcheckCommentSyntaxError: When the comment has invalid syntax
    :return: None
    :yield: Found languages to ignore
    """
    for (index, line) in enumerate(source.splitlines()):
        match = RSTCHECK_COMMENT_REGEX.match(line)
        if match:
            key_and_value = line[match.end() :].strip().split("=")
            if len(key_and_value) != 2:
                raise RstcheckCommentSyntaxError(
                    'Expected "key=value" syntax', line_number=index + 1
                )

            if key_and_value[0].strip() == "ignore-languages":
                for language in key_and_value[1].strip().split(","):
                    yield language.strip()



"""Tests for ``runner`` module."""
# pylint: disable=protected-access
import contextlib
import multiprocessing
import pathlib
import sys
import typing as t

import pytest
import pytest_mock

from rstcheck import checker, config, runner, types
from tests.conftest import EXAMPLES_DIR


class TestRstcheckMainRunnerInit:
    """Test ``RstcheckMainRunner.__init__`` method."""

    @staticmethod
    def test_load_config_file_if_set(mocker: pytest_mock.MockerFixture) -> None:
        """Test config file is loaded if set."""
        mocked_loader = mocker.patch.object(runner.RstcheckMainRunner, "load_config_file")
        config_file_path = pathlib.Path("some-file")
        init_config = config.RstcheckConfig(config_path=config_file_path)

        runner.RstcheckMainRunner([], init_config)  # act

        mocked_loader.assert_called_once_with(config_file_path)

    @staticmethod
    def test_no_load_config_file_if_unset(mocker: pytest_mock.MockerFixture) -> None:
        """Test no config file is loaded if unset."""
        mocked_loader = mocker.patch.object(runner.RstcheckMainRunner, "load_config_file")
        init_config = config.RstcheckConfig()

        runner.RstcheckMainRunner([], init_config)  # act

        mocked_loader.assert_not_called()

    @staticmethod
    @pytest.mark.skipif(sys.platform != "win32", reason="Windows specific.")
    @pytest.mark.parametrize("pool_size", [0, 1, 60, 61, 62, 100])
    def test_max_pool_size_on_windows(pool_size: int, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test pool size is 61 at max on windows."""
        monkeypatch.setattr(multiprocessing, "cpu_count", lambda: pool_size)
        init_config = config.RstcheckConfig()

        result = runner.RstcheckMainRunner([], init_config)._pool_size

        assert result <= 61


class TestRstcheckMainRunnerConfigFileLoader:
    """Test ``RstcheckMainRunner.load_config_file`` method."""

    @staticmethod
    def test_no_config_update_on_no_file_config(monkeypatch: pytest.MonkeyPatch) -> None:
        """Test config is not updated when no file config is found."""
        monkeypatch.setattr(config, "load_config_file_from_path", lambda _: None)
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)

        _runner.load_config_file(pathlib.Path())  # act

        assert _runner.config == init_config

    @staticmethod
    def test_config_update_on_found_file_config(monkeypatch: pytest.MonkeyPatch) -> None:
        """Test config is updated when file config is found."""
        file_config = config.RstcheckConfigFile(report_level=config.ReportLevel.SEVERE)
        monkeypatch.setattr(config, "load_config_file_from_path", lambda _: file_config)
        init_config = config.RstcheckConfig(report_level=config.ReportLevel.INFO)
        _runner = runner.RstcheckMainRunner([], init_config, overwrite_config=True)

        _runner.load_config_file(pathlib.Path())  # act

        assert _runner.config.report_level == config.ReportLevel.SEVERE


class TestRstcheckMainRunnerFileListUpdater:
    """Test ``RstcheckMainRunner.update_file_list`` method."""

    @staticmethod
    def test_empty_file_list() -> None:
        """Test empty file list results in no changes."""
        file_list: t.List[pathlib.Path] = []
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner(file_list, init_config)

        _runner.update_file_list()  # act

        assert not _runner.files_to_check

    @staticmethod
    def test_single_file_in_list() -> None:
        """Test single file in list results in only this file in the list."""
        file_list = [EXAMPLES_DIR / "good" / "rst.rst"]
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner(file_list, init_config)

        _runner.update_file_list()  # act

        assert _runner.files_to_check == file_list

    @staticmethod
    def test_multiple_files_in_list() -> None:
        """Test multiple files in list results in only these files in the list."""
        file_list = [
            EXAMPLES_DIR / "good" / "rst.rst",
            EXAMPLES_DIR / "bad" / "rst.rst",
        ]
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner(file_list, init_config)

        _runner.update_file_list()  # act

        assert _runner.files_to_check == file_list

    @staticmethod
    def test_non_rst_files() -> None:
        """Test non rst files are filtered out."""
        file_list = [
            EXAMPLES_DIR / "good" / "rst.rst",
            EXAMPLES_DIR / "good" / "foo.h",
            EXAMPLES_DIR / "bad" / "rst.rst",
        ]
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner(file_list, init_config)

        _runner.update_file_list()  # act

        assert len(_runner.files_to_check) == 2

    @staticmethod
    def test_directory_without_recursive() -> None:
        """Test directory without recusrive results in empty file list."""
        file_list = [EXAMPLES_DIR / "good"]
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner(file_list, init_config)

        _runner.update_file_list()  # act

        assert not _runner.files_to_check

    @staticmethod
    def test_directory_with_recursive() -> None:
        """Test directory with recusrive results in directories files in file list."""
        file_list = [EXAMPLES_DIR / "good"]
        init_config = config.RstcheckConfig(recursive=True)
        _runner = runner.RstcheckMainRunner(file_list, init_config)

        _runner.update_file_list()  # act

        assert len(_runner.files_to_check) == 6
        assert EXAMPLES_DIR / "good" / "rst.rst" in _runner.files_to_check

    @staticmethod
    def test_dash_as_file() -> None:
        """Test dash as file."""
        file_list = [pathlib.Path("-")]
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner(file_list, init_config)

        _runner.update_file_list()  # act

        assert file_list == _runner.files_to_check

    @staticmethod
    def test_dash_as_file_with_others() -> None:
        """Test dash as file with other files gets ignored."""
        file_list = [pathlib.Path("-"), EXAMPLES_DIR / "good" / "rst.rst"]
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner(file_list, init_config)

        _runner.update_file_list()  # act

        assert len(_runner.files_to_check) == 1
        assert EXAMPLES_DIR / "good" / "rst.rst" in _runner.files_to_check


@pytest.mark.parametrize(
    "lint_errors",
    [[], [types.LintError(source_origin="<string>", line_number=0, message="message")]],
)
def test__run_checks_sync_method(
    lint_errors: t.List[types.LintError], monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test ``RstcheckMainRunner._run_checks_sync`` method.

    Test results are returned.
    """
    monkeypatch.setattr(checker, "check_file", lambda _0, _1, _2: lint_errors)
    file_list = [
        EXAMPLES_DIR / "good" / "rst.rst",
        EXAMPLES_DIR / "bad" / "rst.rst",
    ]
    init_config = config.RstcheckConfig()
    _runner = runner.RstcheckMainRunner(file_list, init_config)

    result = _runner._run_checks_sync()

    assert len(result) == 2
    assert len(result[0]) == len(lint_errors)
    assert len(result[1]) == len(lint_errors)


@pytest.mark.parametrize(
    "lint_errors",
    [[], [types.LintError(source_origin="<string>", line_number=0, message="message")]],
)
def test__run_checks_parallel_method(
    lint_errors: t.List[types.LintError], monkeypatch: pytest.MonkeyPatch
) -> None:  # noqa: AAA05
    """Test ``RstcheckMainRunner._run_checks_parallel`` method.

    Test results are returned.
    The multiprocessing.Pool needs to be mocked, because it interferes with pytest-xdist.
    """

    class MockedPool:  # pylint: disable=too-few-public-methods
        """Mocked instance of ``multiprocessing.Pool``."""

        # noqa: AAA05
        @staticmethod
        def starmap(_0, _1) -> t.List[t.List[types.LintError]]:  # noqa: ANN001
            """Mock for ``multiprocessing.Pool.starmap`` method."""
            return [lint_errors, lint_errors]

    # noqa: AAA05
    @contextlib.contextmanager
    def mock_pool(_) -> t.Generator[MockedPool, None, None]:  # noqa: ANN001
        """Mock context manager for ``multiprocessing.Pool``."""
        yield MockedPool()

    # noqa: AAA05
    monkeypatch.setattr(multiprocessing, "Pool", mock_pool)
    file_list = [
        EXAMPLES_DIR / "good" / "rst.rst",
        EXAMPLES_DIR / "bad" / "rst.rst",
    ]
    init_config = config.RstcheckConfig()
    _runner = runner.RstcheckMainRunner(file_list, init_config)

    result = _runner._run_checks_parallel()  # act

    assert len(result) == 2
    assert len(result[0]) == len(lint_errors)
    assert len(result[1]) == len(lint_errors)


@pytest.mark.parametrize(
    ("results", "error_count"),
    [([], 0), ([[types.LintError(source_origin="<string>", line_number=0, message="message")]], 1)],
)
def test__update_results_method(results: t.List[t.List[types.LintError]], error_count: int) -> None:
    """Test ``RstcheckMainRunner._update_results`` method.

    Test results are set.
    """
    init_config = config.RstcheckConfig()
    _runner = runner.RstcheckMainRunner([], init_config)

    _runner._update_results(results)  # act

    assert len(_runner.errors) == error_count


def test_check_method_sync_with_1_file(mocker: pytest_mock.MockerFixture) -> None:
    """Test ``RstcheckMainRunner.check`` method.

    Test checks are run in sync for 1 file.
    """
    mocked_sync_runner = mocker.patch.object(runner.RstcheckMainRunner, "_run_checks_sync")
    mocked_parallel_runner = mocker.patch.object(runner.RstcheckMainRunner, "_run_checks_parallel")
    init_config = config.RstcheckConfig()
    _runner = runner.RstcheckMainRunner([], init_config)
    _runner.files_to_check = [pathlib.Path("file")]

    _runner.check()  # act

    mocked_sync_runner.assert_called_once()
    mocked_parallel_runner.assert_not_called()


def test_check_method_parallel_with_more_files(mocker: pytest_mock.MockerFixture) -> None:
    """Test ``RstcheckMainRunner.check`` method.

    Test checks are run in parallel for more file.
    """
    mocked_sync_runner = mocker.patch.object(runner.RstcheckMainRunner, "_run_checks_sync")
    mocked_parallel_runner = mocker.patch.object(runner.RstcheckMainRunner, "_run_checks_parallel")
    init_config = config.RstcheckConfig()
    _runner = runner.RstcheckMainRunner([], init_config)
    _runner.files_to_check = [pathlib.Path("file"), pathlib.Path("file2")]

    _runner.check()  # act

    mocked_sync_runner.assert_not_called()
    mocked_parallel_runner.assert_called_once()


class TestRstcheckMainRunnerResultPrinter:
    """Test ``RstcheckMainRunner.get_result`` method."""

    @staticmethod
    def test_exit_code_on_success() -> None:
        """Test exit code 0 is returned on no erros."""
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)

        result = _runner.print_result()

        assert result == 0

    @staticmethod
    def test_success_message_on_success(capsys: pytest.CaptureFixture[str]) -> None:
        """Test success message is printed to stdout by default if no errors."""
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)

        _runner.print_result()  # act

        assert "Success! No issues detected." in capsys.readouterr().out

    @staticmethod
    def test_success_message_print_to_file(tmp_path: pathlib.Path) -> None:
        """Test success message is printed to given file."""
        out_file = tmp_path / "outfile.txt"
        out_file.touch()
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)
        with open(out_file, encoding="utf-8", mode="w") as out_file_handle:

            _runner.print_result(output_file=out_file_handle)  # act

        assert "Success! No issues detected." in out_file.read_text()

    @staticmethod
    def test_exit_code_on_error() -> None:
        """Test exit code 1 is returned when erros were found."""
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)
        _runner.errors = [
            types.LintError(source_origin="<string>", line_number=0, message="message")
        ]

        result = _runner.print_result()

        assert result == 1

    @staticmethod
    def test_no_success_message_on_error(capsys: pytest.CaptureFixture[str]) -> None:
        """Test no succuess message is printed when erros were found."""
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)
        _runner.errors = [
            types.LintError(source_origin="<string>", line_number=0, message="message")
        ]

        _runner.print_result()  # act

        assert "Success! No issues detected." not in capsys.readouterr()

    @staticmethod
    def test_error_printed_to_stderr_by_default(capsys: pytest.CaptureFixture[str]) -> None:
        """Test errors are printed to stderr."""
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)
        _runner._update_results(
            [[types.LintError(source_origin="<string>", line_number=0, message="Some error.")]]
        )

        _runner.print_result()  # act

        assert "(ERROR/3) Some error" in capsys.readouterr().err

    @staticmethod
    def test_error_printed_to_file(tmp_path: pathlib.Path) -> None:
        """Test errors are printed to stderr."""
        out_file = tmp_path / "outfile.txt"
        out_file.touch()
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)
        _runner._update_results(
            [[types.LintError(source_origin="<string>", line_number=0, message="Some error.")]]
        )
        with open(out_file, encoding="utf-8", mode="w") as out_file_handle:

            _runner.print_result(output_file=out_file_handle)  # act

        assert "(ERROR/3) Some error" in out_file.read_text()

    @staticmethod
    def test_error_category_prepend(capsys: pytest.CaptureFixture[str]) -> None:
        """Test ``(ERROR/3)`` is prepended when no category is present."""
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)
        _runner._update_results(
            [[types.LintError(source_origin="<string>", line_number=0, message="Some error.")]]
        )

        _runner.print_result()  # act

        assert "(ERROR/3) Some error." in capsys.readouterr().err

    @staticmethod
    def test_error_message_format(capsys: pytest.CaptureFixture[str]) -> None:
        """Test error message format."""
        init_config = config.RstcheckConfig()
        _runner = runner.RstcheckMainRunner([], init_config)
        _runner._update_results(
            [
                [
                    types.LintError(
                        source_origin="<string>", line_number=0, message="(ERROR/3) Some error."
                    )
                ]
            ]
        )

        _runner.print_result()  # act

        assert "<string>:0: (ERROR/3) Some error." in capsys.readouterr().err



"""CLI for rstcheck."""
import pathlib
import typing as t

import typer

from rstcheck import _compat, _extras, config as config_mod, runner


ValidReportLevels = _compat.Literal["INFO", "WARNING", "ERROR", "SEVERE", "NONE"]

HELP_FILES = "RST files to check. Can be files or directories if --recursive is passed too."
HELP_CONFIG = """Config file to load. Can be a INI file or directory.
If a directory is passed it will be searched for .rstcheck.cfg | setup.cfg.
"""
if _extras.TOMLI_INSTALLED:  # pragma: no cover
    HELP_CONFIG = """Config file to load. Can be a INI or TOML file or directory.
If a directory is passed it will be searched for .rstcheck.cfg | pyproject.toml | setup.cfg.
"""
HELP_RECURSIVE = "Recursively search passed directories for RST files to check."
HELP_REPORT_LEVEL = f"""The report level of the linting issues found.
Valid levels are: INFO | WARNING | ERROR | SEVERE | NONE.
Defauls to {config_mod.DEFAULT_REPORT_LEVEL.name}.
Can be set in config file.
"""
HELP_IGNORE_DIRECTIVES = """Comma-separated-list of directives to add to the ignore list.
Can be set in config file.
"""
HELP_IGNORE_ROLES = """Comma-separated-list of roles to add to the ignore list.
Can be set in config file.
"""
HELP_IGNORE_SUBSTITUTIONS = """Comma-separated-list of substitutions to add to the ignore list.
Can be set in config file.
"""
HELP_IGNORE_LANGUAGES = """Comma-separated-list of languages for code-blocks to add to the ignore
list. The code in ignored code-blocks will not be checked for errors.
Can be set in config file.
"""
HELP_IGNORE_MESSAGES = """A regular expression to match linting issue messages against to ignore.
Can be set in config file.
"""


def cli(  # pylint: disable=too-many-arguments
    files: t.List[pathlib.Path] = typer.Argument(  # noqa: M511,B008
        ..., help=HELP_FILES, allow_dash=True
    ),
    config: t.Optional[pathlib.Path] = typer.Option(  # noqa: M511,B008
        None, "--config", help=HELP_CONFIG
    ),
    recursive: t.Optional[bool] = typer.Option(  # noqa: M511,B008
        None, "--recursive", "-r", help=HELP_RECURSIVE
    ),
    report_level: t.Optional[str] = typer.Option(  # noqa: M511,B008
        None, metavar="LEVEL", help=HELP_REPORT_LEVEL
    ),
    ignore_directives: t.Optional[str] = typer.Option(  # noqa: M511,B008
        None, help=HELP_IGNORE_DIRECTIVES
    ),
    ignore_roles: t.Optional[str] = typer.Option(None, help=HELP_IGNORE_ROLES),  # noqa: M511,B008
    ignore_substitutions: t.Optional[str] = typer.Option(  # noqa: M511,B008
        None, help=HELP_IGNORE_SUBSTITUTIONS
    ),
    ignore_languages: t.Optional[str] = typer.Option(  # noqa: M511,B008
        None, help=HELP_IGNORE_LANGUAGES
    ),
    ignore_messages: t.Optional[str] = typer.Option(  # noqa: M511,B008
        None, metavar="REGEX", help=HELP_IGNORE_MESSAGES
    ),
) -> int:
    """CLI of rstcheck."""
    if pathlib.Path("-") in files and len(files) > 1:
        typer.echo("'-' is only allowed without additional files.", err=True)
        raise typer.Abort()

    rstcheck_config = config_mod.RstcheckConfig(
        config_path=config,
        recursive=recursive,
        report_level=report_level,
        ignore_directives=ignore_directives,
        ignore_roles=ignore_roles,
        ignore_substitutions=ignore_substitutions,
        ignore_languages=ignore_languages,
        ignore_messages=ignore_messages,
    )
    _runner = runner.RstcheckMainRunner(
        check_paths=files, rstcheck_config=rstcheck_config, overwrite_config=False
    )
    _runner.check()
    exit_code = _runner.print_result()
    raise typer.Exit(code=exit_code)


enabled_features = []
if _extras.SPHINX_INSTALLED:
    enabled_features.append("Sphinx")
if _extras.TOMLI_INSTALLED:  # pragma: no cover
    enabled_features.append("Toml")

cli.__doc__ = f"""CLI of rstcheck.

Enabled features: {enabled_features}
"""


def main() -> None:  # pragma: no cover
    """Run CLI."""
    typer.run(cli)


if __name__ == "__main__":
    main()



"""Runner of rstcheck."""
import multiprocessing
import os
import pathlib
import re
import sys
import typing as t

from . import _sphinx, checker, config, types


class RstcheckMainRunner:
    """Main runner of rstcheck."""

    def __init__(
        self,
        check_paths: t.List[pathlib.Path],
        rstcheck_config: config.RstcheckConfig,
        overwrite_config: bool = True,
    ) -> None:
        """Initialize the ``RstcheckMainRunner`` with a base config.

        :param check_paths: Files to check.
        :param rstcheck_config: Base configuration config from e.g. the CLI.
        :param overwrite_config: If file config overwrites current config; defaults to True
        """
        self.config = rstcheck_config
        self.overwrite_config = overwrite_config
        if rstcheck_config.config_path:
            self.load_config_file(rstcheck_config.config_path)

        self.check_paths = check_paths
        self.files_to_check: t.List[pathlib.Path] = []
        self.update_file_list()

        pool_size = multiprocessing.cpu_count()
        # NOTE: Work around https://bugs.python.org/issue45077
        self._pool_size = pool_size if sys.platform != "win32" else min(pool_size, 61)

        self.errors: t.List[types.LintError] = []

    def load_config_file(self, config_path: pathlib.Path) -> None:
        """Load config from file and merge with current config.

        If the loaded file config overwrites the current config depends on the ``overwrite_config``
        attribute set on initialization.

        :param config_path: Path to config file; can be directory or file
        """
        file_config = config.load_config_file_from_path(config_path)

        if file_config is None:
            return

        self.config = config.merge_configs(
            self.config, file_config, config_add_is_dominant=self.overwrite_config
        )

    def update_file_list(self) -> None:  # noqa: CCR001
        """Update file path list with paths specified on initialization.

        Clear the current file list. Then get the file and directory paths specified with
        ``self.check_paths`` and search them for rst files to check. Add those files to the file
        list.
        """
        paths = list(self.check_paths)
        self.files_to_check = []

        if len(paths) == 1 and paths[0].name == "-":
            self.files_to_check.append(paths[0])
            return

        checkable_rst_file: t.Callable[[pathlib.Path], bool] = (
            lambda f: f.is_file() and not f.name.startswith(".") and f.suffix.casefold() == ".rst"
        )

        while paths:
            path = paths.pop(0)
            resolved_path = path.resolve()
            if self.config.recursive and resolved_path.is_dir():
                for root, directories, children in os.walk(path):
                    root_path = pathlib.Path(root)
                    paths += [
                        root_path / f
                        for f in children
                        if checkable_rst_file((root_path / f).resolve())
                    ]
                    directories[:] = [d for d in directories if not d.startswith(".")]
                continue

            if checkable_rst_file(resolved_path) and resolved_path.name != "-":
                self.files_to_check.append(path)

    def _run_checks_sync(self) -> t.List[t.List[types.LintError]]:
        """Check all files from the file list syncronously and return the errors.

        :return: List of lists of errors found per file
        """
        with _sphinx.load_sphinx_if_available():
            results = [
                checker.check_file(file, self.config, self.overwrite_config)
                for file in self.files_to_check
            ]
        return results

    def _run_checks_parallel(self) -> t.List[t.List[types.LintError]]:
        """Check all files from the file list in parallel and return the errors.

        :return: List of lists of errors found per file
        """
        with _sphinx.load_sphinx_if_available(), multiprocessing.Pool(self._pool_size) as pool:
            results = pool.starmap(
                checker.check_file,
                [(file, self.config, self.overwrite_config) for file in self.files_to_check],
            )
        return results

    def _update_results(self, results: t.List[t.List[types.LintError]]) -> None:
        """Take results and update error cache.

        Result normally come from ``self._run_checks_sync`` or ``self._run_checks_parallel``.
        :param results: List of lists of errors found
        """
        self.errors = []
        for errors in results:
            self.errors += errors

    def check(self) -> None:
        """Check all files in the file list and save the errors.

        Multiple files are run in parallel.

        A new call overwrite the old cached errors.
        """
        results = (
            self._run_checks_parallel() if len(self.files_to_check) > 1 else self._run_checks_sync()
        )
        self._update_results(results)

    def print_result(self, output_file: t.Optional[t.TextIO] = None) -> int:
        """Print all cached error messages and return exit code.

        :param output_file: file to print to; defaults to sys.stderr (if ``None``)
        :return: exit code 0 if no error is printed; 1 if any error is printed
        """
        if len(self.errors) == 0:
            print("Success! No issues detected.", file=output_file or sys.stdout)
            return 0

        err_msg_regex = re.compile(r"\([A-Z]+/[0-9]+\)")

        for error in self.errors:
            err_msg = error["message"]
            if not err_msg_regex.match(err_msg):
                err_msg = "(ERROR/3) " + err_msg

            message = f"{error['source_origin']}:{error['line_number']}: {err_msg}"

            print(message, file=output_file or sys.stderr)

        return 1

    def run(self) -> int:  # pragma: no cover
        """Run checks, print error messages and return the result.

        :return: exit code 0 if no error is printed; 1 if any error is printed
        """
        self.check()
        return self.print_result()
